# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.math import Vec2, is_convex_polygon_2d
from ezdxf.math._construct import is_point_in_polygon_2d  # Python version
from ezdxf.acc import USE_C_EXT

is_point_in_polygon_cy = is_point_in_polygon_2d

if USE_C_EXT:
    try:  # Cython version
        from ezdxf.acc.construct import is_point_in_polygon_2d as is_point_in_polygon_cy
    except ImportError:
        pass


def test_inside_horizontal_box():
    square = Vec2.list([(0, 0), (1, 0), (1, 1), (0, 1)])
    assert is_point_in_polygon_2d(Vec2(0.5, 0.5), square) == 1


def test_outside_horizontal_box():
    square = Vec2.list([(0, 0), (1, 0), (1, 1), (0, 1)])
    assert is_point_in_polygon_2d(Vec2(-0.5, 0.5), square) == -1
    assert is_point_in_polygon_2d(Vec2(1.5, 0.5), square) == -1
    assert is_point_in_polygon_2d(Vec2(0.5, -0.5), square) == -1
    assert is_point_in_polygon_2d(Vec2(0.5, 1.5), square) == -1


def test_colinear_outside_horizontal_box():
    square = Vec2.list([(0, 0), (1, 0), (1, 1), (0, 1)])
    assert is_point_in_polygon_2d(Vec2(1.5, 0), square) == -1
    assert is_point_in_polygon_2d(Vec2(-0.5, 0), square) == -1
    assert is_point_in_polygon_2d(Vec2(0, 1.5), square) == -1
    assert is_point_in_polygon_2d(Vec2(0, -0.5), square) == -1


def test_corners_horizontal_box():
    square = Vec2.list([(0, 0), (1, 0), (1, 1), (0, 1)])
    assert is_point_in_polygon_2d(Vec2(0, 0), square) == 0
    assert is_point_in_polygon_2d(Vec2(0, 1), square) == 0
    assert is_point_in_polygon_2d(Vec2(1, 1), square) == 0
    assert is_point_in_polygon_2d(Vec2(0, 1), square) == 0


def test_inside_slanted_box():
    square = Vec2.list([(0, 0), (1, 1), (0, 2), (-1, 1)])
    assert is_point_in_polygon_2d(Vec2(0, 1), square) == 1


def test_outside_slanted_box():
    square = Vec2.list([(0, 0), (1, 1), (0, 2), (-1, 1)])
    assert is_point_in_polygon_2d(Vec2(-1, 0), square) == -1
    assert is_point_in_polygon_2d(Vec2(1, 0), square) == -1
    assert is_point_in_polygon_2d(Vec2(1, 2), square) == -1
    assert is_point_in_polygon_2d(Vec2(-1, 2), square) == -1


def test_corners_slanted_box():
    square = Vec2.list([(0, 0), (1, 1), (0, 2), (-1, 1)])
    assert is_point_in_polygon_2d(Vec2(0, 0), square) == 0
    assert is_point_in_polygon_2d(Vec2(1, 1), square) == 0
    assert is_point_in_polygon_2d(Vec2(0, 2), square) == 0
    assert is_point_in_polygon_2d(Vec2(-1, 1), square) == 0


def test_borders_slanted_box_stable():
    square = Vec2.list([(0, 0), (1, 1), (0, 2), (-1, 1)])
    assert is_point_in_polygon_2d(Vec2(0.5, 0.5), square) == 0
    assert is_point_in_polygon_2d(Vec2(0.5, 1.5), square) == 0
    assert is_point_in_polygon_2d(Vec2(-0.5, 1.5), square) == 0
    assert is_point_in_polygon_2d(Vec2(-0.5, 0.5), square) == 0


# Test concave polygon:
#
#   0123456
# 8 .......
# 7 .7---6.
# 6 .|...|.
# 5 .|.4-5.
# 4 .|.|...
# 3 .|.3-2.
# 2 .|...|.
# 1 .0---1.
# 0 .......
#   0123456

SHAPE_C_CCW = Vec2.list(
    # 0       1       2       3       4       5       6       7
    [(1, 1), (5, 1), (5, 3), (3, 3), (3, 5), (5, 5), (5, 7), (1, 7)]
)
SHAPE_C_CW = list(reversed(SHAPE_C_CCW))

#   0123456
# 8 .......
# 7 .+---+.
# 6 .|678|.
# 5 .|5+-+.
# 4 .|4|...
# 3 .|3+-+.
# 2 .|012|.
# 1 .+---+.
# 0 .......
#   0123456

POINTS_INSIDE = Vec2.list(
    # 0       1       2       3       4       5       6       7       8
    [(2, 2), (3, 2), (4, 2), (2, 3), (2, 4), (2, 5), (2, 6), (3, 6), (4, 6)]
)

#   0123456
# 8 b.c.d.e
# 7 .+---+.
# 6 9|...|a
# 5 .|.+-+.
# 4 6|.|7.8
# 3 .|.+-+.
# 2 4|...|5
# 1 .+---+.
# 0 0.1.2.3
#   0123456

POINTS_OUTSIDE = Vec2.list(
    [
        (0, 0),  # 0
        (2, 0),  # 1
        (4, 0),  # 2
        (6, 0),  # 3
        (0, 2),  # 4
        (6, 2),  # 5
        (0, 4),  # 6
        (4, 4),  # 7
        (6, 2),  # 8
        (0, 6),  # 9
        (6, 6),  # a
        (0, 8),  # b
        (2, 8),  # c
        (4, 8),  # d
        (6, 8),  # e
    ]
)


#   0123456
# 8 .......
# 7 .+-6-+.
# 6 .|...5.
# 5 .|.+4+.
# 4 .7.3...
# 3 .|.+2+.
# 2 .|...1.
# 1 .+-0-+.
# 0 .......
#   0123456

POINTS_ON_BOUNDARY = Vec2.list(
    # 0       1       2       3       4       5       6       7
    [(1, 3), (5, 2), (4, 3), (3, 4), (4, 5), (5, 6), (3, 7), (1, 4)]
)


def test_shape_c_is_not_convex():
    assert is_convex_polygon_2d(SHAPE_C_CCW) is False
    assert is_convex_polygon_2d(SHAPE_C_CW) is False


@pytest.mark.parametrize("point", POINTS_INSIDE)
def test_is_inside_ccw_polygon(point: Vec2):
    assert is_point_in_polygon_2d(point, SHAPE_C_CCW) == 1


@pytest.mark.parametrize("point", POINTS_ON_BOUNDARY)
def test_is_on_boundary_ccw_polygon(point: Vec2):
    assert is_point_in_polygon_2d(point, SHAPE_C_CCW) == 0


@pytest.mark.parametrize("point", POINTS_OUTSIDE)
def test_is_outside_ccw_polygon(point: Vec2):
    assert is_point_in_polygon_2d(point, SHAPE_C_CCW) == -1


@pytest.mark.parametrize("point", POINTS_INSIDE)
def test_is_inside_cw_polygon(point: Vec2):
    assert is_point_in_polygon_2d(point, SHAPE_C_CW) == 1


@pytest.mark.parametrize("point", POINTS_ON_BOUNDARY)
def test_is_on_boundary_cw_polygon(point: Vec2):
    assert is_point_in_polygon_2d(point, SHAPE_C_CW) == 0


@pytest.mark.parametrize("point", POINTS_OUTSIDE)
def test_is_outside_cw_polygon(point: Vec2):
    assert is_point_in_polygon_2d(point, SHAPE_C_CW) == -1


@pytest.mark.skipif(
    is_point_in_polygon_cy is is_point_in_polygon_2d,
    reason="no Cython implementation available",
)
class TestCythonImplementation:
    @pytest.mark.parametrize("point", POINTS_INSIDE)
    def test_is_inside_ccw_polygon(self, point: Vec2):
        assert is_point_in_polygon_cy(point, SHAPE_C_CCW) == 1

    @pytest.mark.parametrize("point", POINTS_ON_BOUNDARY)
    def test_is_on_boundary_ccw_polygon(self, point: Vec2):
        assert is_point_in_polygon_cy(point, SHAPE_C_CCW) == 0

    @pytest.mark.parametrize("point", POINTS_OUTSIDE)
    def test_is_outside_ccw_polygon(self, point: Vec2):
        assert is_point_in_polygon_cy(point, SHAPE_C_CCW) == -1


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf import xref, colors
from ezdxf.document import Drawing


def document_has_no_errors(doc: Drawing) -> bool:
    auditor = doc.audit()
    return not auditor.has_issues


class TestLoadLayers:
    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.filename = "xref.dxf"
        doc.layers.add("Layer0", color=colors.RED)
        doc.layers.add("Layer1", color=colors.GREEN)
        doc.layers.add("Layer2", color=colors.BLUE)
        return doc

    @staticmethod
    def load_layers(sdoc, tdoc, policy):
        loader = xref.Loader(sdoc, tdoc, conflict_policy=policy)
        loader.load_layers(["layer0", "layer1", "layer2"])
        loader.execute()

    def test_conflict_policy_keep(self, sdoc):
        """KEEP: Layer0 of the target doc shouldn't be renamed."""
        tdoc = ezdxf.new()
        tdoc.layers.add("Layer0", color=colors.CYAN)
        self.load_layers(sdoc, tdoc, xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True

        # Layers "0" and "DEFPOINTS" should not be loaded:
        assert len(tdoc.layers) == 3 + 2

        layer0 = tdoc.layers.get("layer0")
        assert layer0.dxf.color == colors.CYAN, "expected Layer0 to be preserved"

    def test_xref_rename_policy(self, sdoc):
        """All layers of the external reference should be renamed.

        Layers "0" and "Defpoints" are always preserved, never renamed.
        """
        tdoc = ezdxf.new()
        tdoc.layers.add("Layer0", color=colors.CYAN)
        self.load_layers(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # Layers "0" and "DEFPOINTS" should not be loaded:
        assert len(tdoc.layers) == 4 + 2

        layer0 = tdoc.layers.get("layer0")
        assert layer0.dxf.color == colors.CYAN, "expected Layer0 to be preserved"

        # check if all loaded layers are renamed:
        xref_layer0 = tdoc.layers.get("xref$0$Layer0")
        assert xref_layer0.dxf.color == colors.RED, "expected loaded Layer0"
        assert xref_layer0.dxf.name == "xref$0$Layer0", "expected updated name"
        assert tdoc.layers.has_entry("xref$0$Layer1"), "expected loaded Layer1"
        assert tdoc.layers.has_entry("xref$0$Layer2"), "expected loaded Layer2"

    def test_xref_rename_policy_load_2_times(self, sdoc):
        """At the 2nd loading process the layer names conflict with the names of the
        1st loading process.
        """
        tdoc = ezdxf.new()
        tdoc.layers.add("Layer0", color=colors.CYAN)
        self.load_layers(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        self.load_layers(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # check if layers of 2nd loading process exist:
        assert tdoc.layers.get("xref$1$Layer0").dxf.name == "xref$1$Layer0"
        assert tdoc.layers.has_entry("xref$1$Layer1") is True
        assert tdoc.layers.has_entry("xref$1$Layer2") is True

    def test_numbered_rename_policy(self, sdoc):
        """Only layers with name conflicts should be renamed.

        Layers "0" and "Defpoints" are always preserved, never renamed.
        """
        tdoc = ezdxf.new()
        tdoc.layers.add("Layer0", color=colors.CYAN)
        self.load_layers(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # Layers "0" and "DEFPOINTS" should not be loaded:
        assert len(tdoc.layers) == 4 + 2

        layer0 = tdoc.layers.get("layer0")
        assert layer0.dxf.color == colors.CYAN, "expected Layer0 to be preserved"

        # check if loaded layers with name conflicts are renamed:
        xref_layer0 = tdoc.layers.get("$0$Layer0")
        assert xref_layer0.dxf.color == colors.RED, "expected loaded Layer0"
        assert xref_layer0.dxf.name == "$0$Layer0", "expected updated name"

        # check if loaded layers without name conflicts preserve their names:
        assert tdoc.layers.has_entry("Layer1") is True
        assert tdoc.layers.has_entry("Layer2") is True

    def test_numbered_rename_policy_load_2_times(self, sdoc):
        """At the 2nd loading process the layer names conflict with the names of the
        1st loading process.
        """
        tdoc = ezdxf.new()
        tdoc.layers.add("Layer0", color=colors.CYAN)
        self.load_layers(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        self.load_layers(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        assert document_has_no_errors(tdoc) is True

        assert tdoc.layers.has_entry("Layer0") is True, "expected preserved Layer0"

        # check renaming schema of 1st loading process:
        assert tdoc.layers.has_entry("$0$Layer0") is True
        assert tdoc.layers.has_entry("Layer1") is True
        assert tdoc.layers.has_entry("Layer2") is True

        # check renaming schema of 2nd loading process:
        assert tdoc.layers.has_entry("$1$Layer0") is True
        assert tdoc.layers.has_entry("$0$Layer1") is True
        assert tdoc.layers.has_entry("$0$Layer2") is True


class TestLoadLinetypes:
    """The implementation in the xref module should be the same as for layers, so these
    tests are as minimal as possible.
    """

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.filename = "xref.dxf"
        doc.linetypes.add("LType0", [0.0], description="xref0")
        doc.linetypes.add("LType1", [0.0], description="xref1")
        return doc

    @staticmethod
    def load_linetypes(sdoc, tdoc, policy):
        loader = xref.Loader(sdoc, tdoc, conflict_policy=policy)
        loader.load_linetypes(["LType0", "LType1"])
        loader.execute()

    def test_conflict_policy_keep(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.linetypes.add("LType0", [0.0], description="preserve")
        self.load_linetypes(sdoc, tdoc, xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True

        # Linetypes "CONTINUOUS", "BYLAYER" and "BYBLOCK" should not be loaded:
        assert len(tdoc.linetypes) == 2 + 3

        ltype0 = tdoc.linetypes.get("LType0")
        assert ltype0.dxf.description == "preserve"
        assert tdoc.linetypes.has_entry("LType1")

    def test_xref_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.linetypes.add("LType0", [0.0], description="preserve")
        self.load_linetypes(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # Linetypes "CONTINUOUS", "BYLAYER" and "BYBLOCK" should not be loaded:
        assert len(tdoc.linetypes) == 3 + 3

        ltype0 = tdoc.linetypes.get("LType0")
        assert ltype0.dxf.description == "preserve"
        assert tdoc.linetypes.get("xref$0$LType0").dxf.name == "xref$0$LType0"
        assert tdoc.linetypes.has_entry("xref$0$LType1") is True

    def test_numbered_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.linetypes.add("LType0", [0.0], description="preserve")
        self.load_linetypes(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # Linetypes "CONTINUOUS", "BYLAYER" and "BYBLOCK" should not be loaded:
        assert len(tdoc.linetypes) == 3 + 3

        ltype0 = tdoc.linetypes.get("LType0")
        assert ltype0.dxf.description == "preserve"
        assert tdoc.linetypes.get("$0$LType0").dxf.name == "$0$LType0"
        assert tdoc.linetypes.has_entry("LType1") is True


class TestLoadTextStyles:
    """The implementation in the xref module should be the same as for layers, so these
    tests are as minimal as possible.
    """

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.filename = "xref.dxf"
        doc.styles.add("Style0", font="style0.ttf")
        doc.styles.add("Style1", font="style1.ttf")
        return doc

    @staticmethod
    def load_styles(sdoc, tdoc, policy):
        loader = xref.Loader(sdoc, tdoc, conflict_policy=policy)
        loader.load_text_styles(["Style0", "Style1", "Standard"])
        loader.execute()

    def test_conflict_policy_keep(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.styles.add("Style0", font="preserve.ttf")
        self.load_styles(sdoc, tdoc, xref.ConflictPolicy.KEEP)

        # style "STANDARD" should not be loaded:
        assert len(tdoc.styles) == 1 + 2

        style0 = tdoc.styles.get("Style0")
        assert style0.dxf.font == "preserve.ttf"
        assert tdoc.styles.has_entry("Style1")

    def test_xref_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.styles.add("Style0", font="preserve.ttf")
        self.load_styles(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # Load 3 text styles from source doc
        assert len(tdoc.styles) == 2 + 3

        style0 = tdoc.styles.get("Style0")
        assert style0.dxf.font == "preserve.ttf"
        assert tdoc.styles.get("xref$0$Style0").dxf.name == "xref$0$Style0"
        assert tdoc.styles.has_entry("xref$0$Style1") is True

        # Text style "Standard" should be loaded and renamed
        assert tdoc.styles.get("xref$0$Standard").dxf.name == "xref$0$Standard"

    def test_numbered_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.styles.add("Style0", font="preserve.ttf")
        self.load_styles(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # Load 3 text styles from source doc
        assert len(tdoc.styles) == 3 + 2

        style0 = tdoc.styles.get("Style0")
        assert style0.dxf.font == "preserve.ttf"
        assert tdoc.styles.get("$0$Style0").dxf.name == "$0$Style0"
        assert tdoc.styles.has_entry("Style1") is True

        # Text style "Standard" should be loaded and renamed
        assert tdoc.styles.get("$0$Standard").dxf.name == "$0$Standard"


class TestLoadDimStyles:
    """The implementation in the xref module should be the same as for layers, so these
    tests are as minimal as possible.
    """

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.filename = "xref.dxf"
        doc.dimstyles.add("Style0", dxfattribs={"dimpost": "xref"})
        doc.dimstyles.add("Style1", dxfattribs={"dimpost": "xref"})
        return doc

    @staticmethod
    def load_styles(sdoc, tdoc, policy):
        loader = xref.Loader(sdoc, tdoc, conflict_policy=policy)
        loader.load_dim_styles(["Style0", "Style1", "Standard"])
        loader.execute()

    def test_conflict_policy_keep(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.dimstyles.add("Style0", dxfattribs={"dimpost": "preserve"})
        self.load_styles(sdoc, tdoc, xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True
        assert len(tdoc.dimstyles) == 1 + 2

        style0 = tdoc.dimstyles.get("Style0")
        assert style0.dxf.dimpost == "preserve"
        assert tdoc.dimstyles.has_entry("Style1")

    def test_xref_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.dimstyles.add("Style0", dxfattribs={"dimpost": "preserve"})
        self.load_styles(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # Load 3 dim styles from source doc
        assert len(tdoc.dimstyles) == 2 + 3

        style0 = tdoc.dimstyles.get("Style0")
        assert style0.dxf.dimpost == "preserve"
        assert tdoc.dimstyles.get("xref$0$Style0").dxf.name == "xref$0$Style0"
        assert tdoc.dimstyles.has_entry("xref$0$Style1") is True

        # Dim style "Standard" should be loaded and renamed
        assert tdoc.dimstyles.get("xref$0$Standard").dxf.name == "xref$0$Standard"

    def test_numbered_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.dimstyles.add("Style0", dxfattribs={"dimpost": "preserve"})
        self.load_styles(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # Load 3 dim styles from source doc
        assert len(tdoc.dimstyles) == 2 + 3

        style0 = tdoc.dimstyles.get("Style0")
        assert style0.dxf.dimpost == "preserve"
        assert tdoc.dimstyles.get("$0$Style0").dxf.name == "$0$Style0"
        assert tdoc.dimstyles.has_entry("Style1") is True

        # Dim style "Standard" should be loaded and renamed
        assert tdoc.dimstyles.get("$0$Standard").dxf.name == "$0$Standard"


class TestLoadMaterials:
    """Materials are stored in object collections which work differently than tables."""

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.filename = "xref.dxf"
        doc.materials.new("Mat0").dxf.description = "XrefMat0"
        doc.materials.new("Mat1").dxf.description = "XrefMat1"
        return doc

    @staticmethod
    def load_materials(sdoc, tdoc, policy):
        loader = xref.Loader(sdoc, tdoc, conflict_policy=policy)
        loader.load_materials(["mat0", "mat1"])
        loader.execute()

    def test_conflict_policy_keep(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.materials.new("Mat0").dxf.description = "preserve"
        self.load_materials(sdoc, tdoc, xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True

        # material "GLOBAL", "BYLAYER" and "BYBLOCK" should not be loaded:
        assert len(tdoc.materials) == 3 + 2

        mat0 = tdoc.materials.get("Mat0")
        assert mat0.dxf.description == "preserve"
        assert tdoc.materials.has_entry("Mat1") is True

    def test_xref_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.materials.new("Mat0").dxf.description = "preserve"
        self.load_materials(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # material "GLOBAL", "BYLAYER" and "BYBLOCK" should not be loaded:
        assert len(tdoc.materials) == 3 + 3

        mat0 = tdoc.materials.get("Mat0")
        assert mat0.dxf.description == "preserve"
        assert tdoc.materials.get("xref$0$Mat0").dxf.name == "xref$0$Mat0"
        assert tdoc.materials.has_entry("xref$0$Mat1") is True

    def test_numbered_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.materials.new("Mat0").dxf.description = "preserve"
        self.load_materials(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # material "GLOBAL", "BYLAYER" and "BYBLOCK" should not be loaded:
        assert len(tdoc.materials) == 3 + 3

        mat0 = tdoc.materials.get("Mat0")
        assert mat0.dxf.description == "preserve"
        assert tdoc.materials.get("$0$Mat0").dxf.name == "$0$Mat0"
        assert tdoc.materials.has_entry("Mat1") is True


class TestLoadMLineStyles:
    """MLineStyles are stored in object collections like Materials."""

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.filename = "xref.dxf"
        mline_style0 = doc.mline_styles.new("Style0")
        mline_style0.dxf.description = "XrefStyle0"
        mline_style0.elements.append(1)  # type: ignore

        mline_style1 = doc.mline_styles.new("Style1")
        mline_style1.dxf.description = "XrefStyle1"
        mline_style1.elements.append(1)  # type: ignore
        return doc

    @staticmethod
    def load_mline_styles(sdoc, tdoc, policy):
        loader = xref.Loader(sdoc, tdoc, conflict_policy=policy)
        loader.load_mline_styles(["style0", "style1"])
        loader.execute()

    def test_conflict_policy_keep(self, sdoc):
        tdoc = ezdxf.new()
        mline_style0 = tdoc.mline_styles.new("Style0")
        mline_style0.dxf.description = "preserve"
        mline_style0.elements.append(1)  # type: ignore

        self.load_mline_styles(sdoc, tdoc, xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True

        # style "Standard" should not be loaded:
        assert len(tdoc.mline_styles) == 1 + 2

        style0 = tdoc.mline_styles.get("Style0")
        assert style0.dxf.description == "preserve"
        assert tdoc.mline_styles.has_entry("Style1") is True

    def test_xref_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        mline_style0 = tdoc.mline_styles.new("Style0")
        mline_style0.dxf.description = "preserve"
        mline_style0.elements.append(1)  # type: ignore

        self.load_mline_styles(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # style "Standard" should not be loaded:
        assert len(tdoc.mline_styles) == 1 + 3

        style0 = tdoc.mline_styles.get("Style0")
        assert style0.dxf.description == "preserve"
        assert tdoc.mline_styles.get("xref$0$Style0").dxf.name == "xref$0$Style0"
        assert tdoc.mline_styles.has_entry("xref$0$Style1") is True

    def test_numbered_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        mline_style0 = tdoc.mline_styles.new("Style0")
        mline_style0.dxf.description = "preserve"
        mline_style0.elements.append(1)  # type: ignore

        self.load_mline_styles(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # style "Standard" should not be loaded:
        assert len(tdoc.mline_styles) == 1 + 3

        style0 = tdoc.mline_styles.get("Style0")
        assert style0.dxf.description == "preserve"
        assert tdoc.mline_styles.get("$0$Style0").dxf.name == "$0$Style0"
        assert tdoc.mline_styles.has_entry("Style1") is True


class TestLoadMLeaderStyles:
    """MLeaderStyles are stored in object collections like Materials."""

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.filename = "xref.dxf"
        doc.mleader_styles.new("Style0").dxf.default_text_content = "XrefStyle0"
        doc.mleader_styles.new("Style1").dxf.default_text_content = "XrefStyle1"
        return doc

    @staticmethod
    def load_mleader_styles(sdoc, tdoc, policy):
        loader = xref.Loader(sdoc, tdoc, conflict_policy=policy)
        loader.load_mleader_styles(["style0", "style1"])
        loader.execute()

    def test_conflict_policy_keep(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.mleader_styles.new("Style0").dxf.default_text_content = "preserve"
        self.load_mleader_styles(sdoc, tdoc, xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True

        # style "Standard" should not be loaded:
        assert len(tdoc.mleader_styles) == 1 + 2

        style0 = tdoc.mleader_styles.get("Style0")
        assert style0.dxf.default_text_content == "preserve"
        assert tdoc.mleader_styles.has_entry("Style1") is True

    def test_xref_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.mleader_styles.new("Style0").dxf.default_text_content = "preserve"
        self.load_mleader_styles(sdoc, tdoc, xref.ConflictPolicy.XREF_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # style "Standard" should not be loaded:
        assert len(tdoc.mleader_styles) == 1 + 3

        style0 = tdoc.mleader_styles.get("Style0")
        assert style0.dxf.default_text_content == "preserve"
        assert tdoc.mleader_styles.get("xref$0$Style0").dxf.name == "xref$0$Style0"
        assert tdoc.mleader_styles.has_entry("xref$0$Style1") is True

    def test_numbered_rename_policy(self, sdoc):
        tdoc = ezdxf.new()
        tdoc.mleader_styles.new("Style0").dxf.default_text_content = "preserve"
        self.load_mleader_styles(sdoc, tdoc, xref.ConflictPolicy.NUM_PREFIX)
        assert document_has_no_errors(tdoc) is True

        # style "Standard" should not be loaded:
        assert len(tdoc.mleader_styles) == 1 + 3

        style0 = tdoc.mleader_styles.get("Style0")
        assert style0.dxf.default_text_content == "preserve"
        assert tdoc.mleader_styles.get("$0$Style0").dxf.name == "$0$Style0"
        assert tdoc.mleader_styles.has_entry("Style1") is True


class TestLoadTextWithExistingTextstyle:
    def make_doc(self, font: str) -> Drawing:
        doc = ezdxf.new()
        doc.styles.add("Style0", font=font)
        msp = doc.modelspace()
        msp.add_text("MyText", dxfattribs={"style": "Style0"})
        return doc

    def test_keep_policy(self):
        sdoc = self.make_doc("arial.ttf")
        tdoc = self.make_doc("times.ttf")

        xref.load_modelspace(sdoc, tdoc, conflict_policy=xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True

        assert len(tdoc.styles) == 2
        assert tdoc.styles.get("Style0").dxf.font == "times.ttf"

        existing_text, loaded_text = tdoc.modelspace()
        assert existing_text.dxf.style == "Style0"
        assert loaded_text.dxf.style == "Style0"

    def test_xref_policy(self):
        sdoc = self.make_doc("arial.ttf")
        sdoc.filename = "xref.dxf"
        tdoc = self.make_doc("times.ttf")

        xref.load_modelspace(
            sdoc, tdoc, conflict_policy=xref.ConflictPolicy.XREF_PREFIX
        )
        assert document_has_no_errors(tdoc) is True

        assert len(tdoc.styles) == 3  # includes "Standard" style
        assert tdoc.styles.get("Style0").dxf.font == "times.ttf"
        assert tdoc.styles.get("xref$0$Style0").dxf.font == "arial.ttf"

        existing_text, loaded_text = tdoc.modelspace()
        assert existing_text.dxf.style == "Style0", "expected style for tdoc"
        assert loaded_text.dxf.style == "xref$0$Style0", "expected mapped style name"


class TestLoadLineWithExistingLinetype:
    def make_doc(self, desc: str) -> Drawing:
        doc = ezdxf.new()
        doc.linetypes.add("LType0", [0.0], description=desc)
        msp = doc.modelspace()
        msp.add_line((0, 0), (1, 0), dxfattribs={"linetype": "LType0"})
        return doc

    def test_keep_policy(self):
        sdoc = self.make_doc("source")
        tdoc = self.make_doc("keep")

        xref.load_modelspace(sdoc, tdoc, conflict_policy=xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True

        assert len(tdoc.linetypes) == 4  # incl. CONTINUOUS, BYLAYER, BYBLOCK
        assert tdoc.linetypes.get("LType0").dxf.description == "keep"

        existing_line, loaded_line = tdoc.modelspace()
        assert existing_line.dxf.linetype == "LType0"
        assert loaded_line.dxf.linetype == "LType0"

    def test_xref_policy(self):
        sdoc = self.make_doc("source")
        sdoc.filename = "xref.dxf"
        tdoc = self.make_doc("keep")

        xref.load_modelspace(
            sdoc, tdoc, conflict_policy=xref.ConflictPolicy.XREF_PREFIX
        )
        assert document_has_no_errors(tdoc) is True

        assert len(tdoc.linetypes) == 5  # incl. CONTINUOUS, BYLAYER, BYBLOCK
        assert tdoc.linetypes.get("LType0").dxf.description == "keep"
        assert tdoc.linetypes.get("xref$0$LType0").dxf.description == "source"

        existing_line, loaded_line = tdoc.modelspace()
        assert existing_line.dxf.linetype == "LType0", "expected style for tdoc"
        assert (
            loaded_line.dxf.linetype == "xref$0$LType0"
        ), "expected mapped linetype name"


class TestLoadLineWithExistingComplexLinetype:
    def make_doc(self) -> Drawing:
        doc = ezdxf.new()
        doc.styles.add_shx("ltypeshp.shx")
        doc.linetypes.add(  # see also: complex_line_type_example.py
            "SQUARE",
            pattern="A,.25,-.1,[132,ltypeshp.shx,x=-.1,s=.1],-.1,1",
            description="Square -[]-----[]-----[]-----[]----[]----",
            length=1.45,
        )
        msp = doc.modelspace()
        msp.add_line((0, 0), (1, 0), dxfattribs={"linetype": "SQUARE"})
        return doc

    def test_xref_policy(self):
        """Shape files do not get renamed at loading process.

        The style handle of the loaded linetype 'xref$0$SQUARE' should be replaced by
        the handle of the existing shape file 'ltypeshp.shx' in tdoc.
        """
        sdoc = self.make_doc()
        sdoc.filename = "xref.dxf"

        tdoc = self.make_doc()
        existing_shape_file = tdoc.styles.find_shx("ltypeshp.shx")
        xref.load_modelspace(
            sdoc, tdoc, conflict_policy=xref.ConflictPolicy.XREF_PREFIX
        )
        assert document_has_no_errors(tdoc) is True

        shape_file = tdoc.styles.find_shx("ltypeshp.shx")
        assert shape_file is existing_shape_file
        loaded_ltype = tdoc.linetypes.get("xref$0$SQUARE")
        assert (
            loaded_ltype.pattern_tags.get_style_handle() == shape_file.dxf.handle
        ), "expected handle is mapped to existing shape file"


class TestLoadEntityWithExistingMaterial:
    def make_doc(self, desc: str) -> Drawing:
        doc = ezdxf.new()
        mat0 = doc.materials.new("Mat0")
        mat0.dxf.description = desc
        msp = doc.modelspace()
        msp.add_line((0, 0), (1, 0), dxfattribs={"material_handle": mat0.dxf.handle})
        return doc

    def test_keep_policy(self):
        """The material_handle of the loaded entity should be the handle of the existing
        material "Mat0" in tdoc
        """
        sdoc = self.make_doc("source")
        tdoc = self.make_doc("keep")

        xref.load_modelspace(sdoc, tdoc, conflict_policy=xref.ConflictPolicy.KEEP)
        assert document_has_no_errors(tdoc) is True

        assert len(tdoc.materials) == 4  # incl. GLOBAL, BYLAYER, BYBLOCK
        mat0 = tdoc.materials.get("Mat0")
        assert mat0.dxf.description == "keep"
        existing_line, loaded_line = tdoc.modelspace()
        assert existing_line.dxf.material_handle == mat0.dxf.handle
        assert loaded_line.dxf.material_handle == mat0.dxf.handle

    def test_xref_policy(self):
        """The material_handle of the loaded entity should be the handle of the loaded
        material "xref$0$Mat0".
        """
        sdoc = self.make_doc("source")
        sdoc.filename = "xref.dxf"
        tdoc = self.make_doc("keep")

        xref.load_modelspace(
            sdoc, tdoc, conflict_policy=xref.ConflictPolicy.XREF_PREFIX
        )
        assert document_has_no_errors(tdoc) is True
        assert len(tdoc.materials) == 5  # incl. GLOBAL, BYLAYER, BYBLOCK

        existing_mat0 = tdoc.materials.get("Mat0")
        loaded_mat0 = tdoc.materials.get("xref$0$Mat0")
        assert loaded_mat0.dxf.name == "xref$0$Mat0"
        assert loaded_mat0.dxf.description == "source"

        existing_line, loaded_line = tdoc.modelspace()
        assert existing_line.dxf.material_handle == existing_mat0.dxf.handle
        assert loaded_line.dxf.material_handle == loaded_mat0.dxf.handle


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.render import R12Spline


CONTROL_POINTS = [
    (8.55, 2.96),
    (8.55, -0.03),
    (2.75, -0.03),
    (2.76, 3.05),
    (4.29, 1.78),
    (6.79, 3.05),
]


@pytest.fixture(scope="module")
def msp():
    return ezdxf.new("R12").modelspace()


def test_r12_quadratic_spline(msp):
    spline = R12Spline(CONTROL_POINTS, degree=2, closed=False)
    polyline = spline.render(msp, segments=40)
    assert polyline.dxftype() == "POLYLINE"
    assert polyline.dxf.layer == "0"
    assert len(polyline) == 41 + len(CONTROL_POINTS)
    assert polyline.is_closed is False


def test_r12_cubic_spline(msp):
    spline = R12Spline(CONTROL_POINTS, degree=3, closed=False)
    polyline = spline.render(msp, segments=40)
    assert polyline.dxftype() == "POLYLINE"
    assert polyline.dxf.layer == "0"
    assert len(polyline) == 41 + len(CONTROL_POINTS)
    assert polyline.is_closed is False


def test_r12_cubic_spline_closed(msp):
    spline = R12Spline(CONTROL_POINTS, degree=3, closed=True)
    polyline = spline.render(msp, segments=40)
    assert polyline.dxftype() == "POLYLINE"
    assert polyline.dxf.layer == "0"
    assert len(polyline) == 41 + len(CONTROL_POINTS)
    assert polyline.is_closed is True



# Copyright (c) 2011-2022, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

from ezdxf.entities import (
    Dictionary,
    DictionaryWithDefault,
    DictionaryVar,
    DXFEntity,
    factory,
    DXFGraphic,
    XRecord,
)

from ezdxf import DXFKeyError, DXFValueError
from ezdxf.audit import Auditor, AuditError


class MockDoc:
    class Objects:
        def delete_entity(self, entity):
            pass

    class EntityDB:
        def __getitem__(self, item):
            return item

    def __init__(self):
        self.entitydb = MockDoc.EntityDB()
        self.dxfversion = "AC1015"
        self.objects = MockDoc.Objects()


class TestNoneEmptyDXFDict:
    @pytest.fixture
    def dxfdict(self):
        d = Dictionary.from_text(ROOTDICT, doc=MockDoc())
        return d

    def test_getitem(self, dxfdict):
        assert dxfdict["ACAD_PLOTSTYLENAME"] == "E"

    def test_len(self, dxfdict):
        assert 14 == len(dxfdict)

    def test_getitem_with_keyerror(self, dxfdict):
        with pytest.raises(ezdxf.DXFKeyError):
            _ = dxfdict["MOZMAN"]

    def test_owner(self, dxfdict):
        assert dxfdict.dxf.owner == "0"

    def test_handle(self, dxfdict):
        assert dxfdict.dxf.handle == "C"

    def test_get(self, dxfdict):
        assert dxfdict.get("ACAD_MOZMAN", "XXX") == "XXX"

    def test_get_entity(self, dxfdict):
        # returns just the handle, because no associated drawing exists
        assert "E" == dxfdict.get("ACAD_PLOTSTYLENAME")

    def test_get_without_keyerror(self, dxfdict):
        assert dxfdict.get("ACAD_MOZMAN") is None

    def test_contains(self, dxfdict):
        assert "ACAD_PLOTSTYLENAME" in dxfdict

    def test_not_contains(self, dxfdict):
        assert "MOZMAN" not in dxfdict

    def test_delete_existing_key(self, dxfdict):
        del dxfdict["ACAD_PLOTSTYLENAME"]
        assert "ACAD_PLOTSTYLENAME" not in dxfdict
        assert 13 == len(dxfdict)

    def test_delete_not_existing_key(self, dxfdict):
        with pytest.raises(ezdxf.DXFKeyError):
            del dxfdict["MOZMAN"]

    def test_remove_existing_key(self, dxfdict):
        dxfdict.remove("ACAD_PLOTSTYLENAME")
        assert "ACAD_PLOTSTYLENAME" not in dxfdict
        assert 13 == len(dxfdict)

    def test_remove_not_existing_key(self, dxfdict):
        with pytest.raises(ezdxf.DXFKeyError):
            dxfdict.remove("MOZMAN")

    def test_discard_existing_key(self, dxfdict):
        dxfdict.discard("ACAD_PLOTSTYLENAME")
        assert "ACAD_PLOTSTYLENAME" not in dxfdict
        assert 13 == len(dxfdict)

    def test_discard_not_existing_key(self, dxfdict):
        dxfdict.discard("MOZMAN")
        assert 14 == len(dxfdict)

    def test_clear(self, dxfdict):
        assert 14 == len(dxfdict)
        dxfdict.clear()
        assert 0 == len(dxfdict)

    def test_keys(self, dxfdict):
        assert 14 == len(list(dxfdict.keys()))

    def test_items(self, dxfdict):
        assert 14 == len(list(dxfdict.items()))

    def test_find_key(self, dxfdict):
        entry = dxfdict["ACAD_COLOR"]
        assert dxfdict.find_key(entry) == "ACAD_COLOR"


class TestEmptyDXFDict:
    @pytest.fixture
    def dxfdict(self):
        return Dictionary.from_text(EMPTY_DICT, doc=MockDoc())

    def test_len(self, dxfdict):
        assert 0 == len(dxfdict)

    def test_add_first_item(self, dxfdict):
        dxfdict["TEST"] = "ABBA"
        assert 1 == len(dxfdict)
        assert "ABBA" == dxfdict["TEST"]

    def test_add_first_item_2(self, dxfdict):
        dxfdict.add("TEST", "ABBA")
        assert 1 == len(dxfdict)
        assert "ABBA" == dxfdict["TEST"]

    def test_add_and_replace_first_item(self, dxfdict):
        dxfdict["TEST"] = "ABBA"
        assert 1 == len(dxfdict)
        assert "ABBA" == dxfdict["TEST"]
        dxfdict["TEST"] = "FEFE"
        assert 1 == len(dxfdict)
        assert "FEFE" == dxfdict["TEST"]

    def test_clear(self, dxfdict):
        assert 0 == len(dxfdict)
        dxfdict.clear()
        assert 0 == len(dxfdict)


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_do_not_recover_owner_tag_at_loading_stage():
    # this is the task of the audit process and breaks otherwise the
    # recover mode for test file: "ProE_AC1018.dxf"
    d = Dictionary.from_text(WITHOUT_OWNER_TAG)
    assert d.dxf.hasattr("owner") is False


def test_dictionary_loads_owner_tag():
    d = Dictionary.from_text(WITH_OWNER_TAG)
    assert d.dxf.owner == "FEFE"


def test_get_entity_invalid_handle(doc):
    rootdict = doc.rootdict
    e = factory.new("ACDBPLACEHOLDER", {})
    e.dxf.handle = "FFFF"
    e.dxf.owner = "ABBA"
    rootdict["TEST"] = e
    assert rootdict["TEST"].dxf.handle == "FFFF"
    assert rootdict["TEST"].dxf.owner == "ABBA"

    with pytest.raises(DXFValueError):
        rootdict["TEST"] = "XXX"

    with pytest.raises(DXFKeyError):
        _ = rootdict["MOZMAN"]

    dxfdict = doc.rootdict.get_required_dict("TEST2")
    assert (
        dxfdict.dxftype() == "DICTIONARY"
    ), "previous statement should not raise an exception"


def test_add_sub_dict(doc):
    rootdict = doc.rootdict
    assert "MOZMAN_TEST" not in rootdict
    new_dict = rootdict.get_required_dict("MOZMAN_TEST")
    assert "DICTIONARY" == new_dict.dxftype()
    assert "MOZMAN_TEST" in rootdict
    assert new_dict.dxf.owner == rootdict.dxf.handle


class TestDictionaryVar:
    def test_add_dict_var(self, doc):
        rootdict = doc.rootdict
        assert "MOZMAN_VAR" not in rootdict
        new_var = rootdict.add_dict_var("MOZMAN_VAR", "Hallo")
        assert new_var.dxftype() == "DICTIONARYVAR"
        assert "MOZMAN_VAR" in rootdict
        assert new_var.dxf.value == "Hallo"
        assert new_var.dxf.owner == rootdict.dxf.handle

    def test_value_property(self, doc):
        rootdict = doc.rootdict
        new_var = rootdict.add_dict_var("MOZMAN_VAR2", "")
        assert new_var.value == ""
        new_var.value = "Hello"
        assert new_var.value == "Hello"


def test_add_xrecord(doc):
    rootdict = doc.rootdict
    assert "MOZMAN_XRECORD" not in rootdict
    xrecord = rootdict.add_xrecord("MOZMAN_XRECORD")
    assert xrecord.dxftype() == "XRECORD"
    assert "MOZMAN_XRECORD" in rootdict
    assert xrecord.dxf.owner == rootdict.dxf.handle


def test_cannot_add_graphical_entities_to_dict(doc):
    line = doc.modelspace().add_line((0, 0), (10, 0))
    assert isinstance(line, DXFGraphic)
    with pytest.raises(ezdxf.DXFTypeError):
        doc.rootdict["LINE"] = line


def test_audit_fix_invalid_root_dict_owner():
    doc = ezdxf.new()
    rootdict = doc.rootdict
    auditor = Auditor(doc)

    rootdict.dxf.owner = "FF"  # set invalid owner
    auditor.run()
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_OWNER_HANDLE
    assert rootdict.dxf.owner == "0"


def test_audit_ok():
    doc = ezdxf.new()
    auditor = Auditor(doc)

    rootdict = doc.rootdict
    assert "TEST_VAR_1" not in rootdict
    new_var = rootdict.add_dict_var("TEST_VAR_1", "Hallo")
    assert new_var.dxftype() == "DICTIONARYVAR"

    rootdict.audit(auditor)
    assert len(auditor) == 0


def test_audit_invalid_pointer():
    doc = ezdxf.new()
    auditor = Auditor(doc)

    d = doc.rootdict.add_new_dict("TEST_AUDIT_2")
    entry = d.add_dict_var("TEST_VAR_2", "Hallo")
    entry.dxf.handle = "ABBA"
    d.audit(auditor)
    assert len(auditor.fixes) == 1
    e = auditor.fixes[0]
    assert e.code == AuditError.INVALID_DICTIONARY_ENTRY


def test_audit_fix_invalid_pointer():
    doc = ezdxf.new()
    auditor = Auditor(doc)

    d = doc.rootdict.add_new_dict("TEST_AUDIT_3")
    entry = d.add_dict_var("TEST_VAR_3", "Hallo")
    entry.dxf.handle = "ABBA"
    d.audit(auditor)
    assert len(auditor) == 0, "should return count of unfixed errors"
    assert len(auditor.errors) == 0
    assert len(auditor.fixes) == 1
    fix = auditor.fixes[0]
    assert fix.code == AuditError.INVALID_DICTIONARY_ENTRY

    # test if invalid entry was removed
    assert len(d) == 0
    assert "TEST_VAR_3" not in d


def test_link_dxf_object_to_dictionary():
    from ezdxf.entities import DXFObject

    dictionary = Dictionary.new(handle="ABBA")
    obj = DXFObject.new(handle="FEFE")
    dictionary.link_dxf_object("MyEntry", obj)
    assert "MyEntry" in dictionary
    assert obj.dxf.owner == "ABBA"


class TestDXFDictWithDefault:
    @pytest.fixture
    def dxfdict(self):
        d = DictionaryWithDefault.from_text(DEFAULT_DICT, doc=MockDoc())
        d._default = DXFEntity()
        d._default.dxf.handle = "FEFE"
        return d

    def test_get_existing_value(self, dxfdict):
        assert "F" == dxfdict["Normal"]

    def test_get_not_existing_value(self, dxfdict):
        assert dxfdict["Mozman"].dxf.handle == "FEFE"

    def test_get_default_value(self, dxfdict):
        assert "F" == dxfdict.dxf.default

    def test_set_default_value(self, dxfdict):
        e = DXFEntity()
        e.dxf.handle = "ABBA"

        dxfdict.set_default(e)
        assert dxfdict["Mozman"].dxf.handle == "ABBA"

    def test_create_place_holder_for_invalid_default_vaue(self):
        doc = ezdxf.new()
        d = doc.objects.add_dictionary_with_default(
            owner=doc.rootdict.dxf.handle, default="0"
        )
        auditor = Auditor(doc)
        d.audit(auditor)
        default = d.get("xxx")
        assert default.is_alive is True
        assert default.dxftype() == "ACDBPLACEHOLDER"


class TestCopyHardOwnerDictionary:
    @pytest.fixture(scope="class")
    def source(self, doc) -> Dictionary:
        doc = ezdxf.new()
        dictionary = doc.rootdict.get_required_dict("COPYTEST", hard_owned=True)
        dictionary.add_dict_var("DICTVAR", "VarContent")
        xrecord = dictionary.add_xrecord("XRECORD")
        xrecord.reset([(1, "String"), (40, 3.1415), (90, 4711)])
        return dictionary

    def test_keys_are_copied(self, source: Dictionary):
        copy = source.copy()
        assert "DICTVAR" in copy
        assert "XRECORD" in copy

    def test_objects_exist(self, source: Dictionary):
        copy = source.copy()
        assert isinstance(copy["DICTVAR"], DictionaryVar)
        assert isinstance(copy["XRECORD"], XRecord)

    def test_objects_are_copied(self, source: Dictionary):
        copy = source.copy()
        assert copy["DICTVAR"] is not source["DICTVAR"]
        assert copy["XRECORD"] is not source["XRECORD"]

    def test_copied_objects_are_not_bound_to_document(self, source: Dictionary):
        copy = source.copy()
        assert factory.is_bound(copy["DICTVAR"], source.doc) is False
        assert factory.is_bound(copy["XRECORD"], source.doc) is False

    def test_fully_manual_dictionary_copy(self, source: Dictionary):
        doc = source.doc
        # manual copy procedure:
        copy = source.copy()
        factory.bind(copy, doc)
        doc.objects.add_object(copy)
        # this is all done automatically if you use:
        # doc.entitydb.duplicate_entity(source)
        assert copy in doc.objects
        assert factory.is_bound(copy, doc)

    @pytest.mark.parametrize("key", ["DICTVAR", "XRECORD"])
    def test_copied_data_is_bound(self, source: Dictionary, key: str):
        doc = source.doc
        copy = doc.entitydb.duplicate_entity(source)
        source_handle = source[key].dxf.handle
        copied_handle = copy[key].dxf.handle
        assert copied_handle is not None
        assert source_handle != copied_handle
        assert copied_handle in doc.entitydb
        assert copy[key] in doc.objects
        assert factory.is_bound(copy[key], doc)

    def test_copied_xrecord(self, source: Dictionary):
        doc = source.doc
        copy = doc.entitydb.duplicate_entity(source)
        xrecord = copy["XRECORD"]
        assert xrecord.tags == [(1, "String"), (40, 3.1415), (90, 4711)]

    def test_copied_objects_are_owned_by_copy(self, source: Dictionary):
        copy = source.copy()
        assert copy["DICTVAR"].dxf.owner == copy.dxf.handle
        assert copy["XRECORD"].dxf.owner == copy.dxf.handle

    def test_copy_sub_dictionary(self, source: Dictionary):
        doc = source.doc
        source = doc.entitydb.duplicate_entity(source)
        sub_dict = source.add_new_dict("SUBDICT", hard_owned=True)
        sub_dict.add_dict_var("SUBDICTVAR", "SubVarContent")
        copy = doc.entitydb.duplicate_entity(source)

        sub_dict_copy = copy["SUBDICT"]
        assert sub_dict is not sub_dict_copy
        assert sub_dict["SUBDICTVAR"] is not sub_dict_copy["SUBDICTVAR"]
        assert factory.is_bound(sub_dict_copy, doc) is True
        assert factory.is_bound(sub_dict_copy["SUBDICTVAR"], doc) is True


class TestCopyNotHardOwnerDictionary:
    @pytest.fixture(scope="class")
    def source(self, doc) -> Dictionary:
        doc = ezdxf.new()
        dictionary = doc.rootdict.get_required_dict("COPYTEST", hard_owned=True)
        dict_var = dictionary.add_dict_var("DICTVAR", "VarContent")
        xrecord = dictionary.add_xrecord("XRECORD")
        xrecord.reset([(1, "String"), (40, 3.1415), (90, 4711)])
        dictionary2 = doc.rootdict.get_required_dict("COPYTEST2", hard_owned=False)
        dictionary2["DICTVAR"] = dict_var
        dictionary2["XRECORD"] = xrecord
        return dictionary2

    def test_is_not_hard_owner(self, source: Dictionary):
        assert source.is_hard_owner is False

    def test_keys_are_copied(self, source: Dictionary):
        copy = source.copy()
        assert "DICTVAR" in copy
        assert "XRECORD" in copy

    def test_objects_exist(self, source: Dictionary):
        copy = source.copy()
        assert isinstance(copy["DICTVAR"], DictionaryVar)
        assert isinstance(copy["XRECORD"], XRecord)

    def test_objects_are_not_copied(self, source: Dictionary):
        copy = source.copy()
        assert copy["DICTVAR"] is source["DICTVAR"]
        assert copy["XRECORD"] is source["XRECORD"]

    def test_objects_are_not_owned_by_copy(self, source: Dictionary):
        copy = source.copy()
        assert copy["DICTVAR"].dxf.owner != copy.dxf.handle
        assert copy["XRECORD"].dxf.owner != copy.dxf.handle

    def test_copied_dictionary_is_bound(self, source: Dictionary):
        doc = source.doc
        copy = doc.entitydb.duplicate_entity(source)
        assert copy in doc.objects
        assert factory.is_bound(copy, doc)


ROOTDICT = """0
DICTIONARY
5
C
330
0
100
AcDbDictionary
281
1
3
ACAD_COLOR
350
73
3
ACAD_GROUP
350
D
3
ACAD_LAYOUT
350
1A
3
ACAD_MATERIAL
350
72
3
ACAD_MLEADERSTYLE
350
D7
3
ACAD_MLINESTYLE
350
17
3
ACAD_PLOTSETTINGS
350
19
3
ACAD_PLOTSTYLENAME
350
E
3
ACAD_SCALELIST
350
B6
3
ACAD_TABLESTYLE
350
86
3
ACAD_VISUALSTYLE
350
99
3
ACDB_RECOMPOSE_DATA
350
40F
3
AcDbVariableDictionary
350
66
3
DWGPROPS
350
410
"""

EMPTY_DICT = """  0
DICTIONARY
  5
C
330
0
100
AcDbDictionary
281
     1
"""

DEFAULT_DICT = """  0
ACDBDICTIONARYWDFLT
  5
E
102
{ACAD_REACTORS
330
C
102
}
330
C
100
AcDbDictionary
281
     1
  3
Normal
350
F
100
AcDbDictionaryWithDefault
340
F
"""

WITHOUT_OWNER_TAG = """0
DICTIONARY
5
C
100
AcDbDictionary
281
1
"""

WITH_OWNER_TAG = """0
DICTIONARY
5
C
330
FEFE
100
AcDbDictionary
281
1
"""



# Copyright (c) 2018-2022 Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from math import radians
from ezdxf.audit import Auditor
from ezdxf.layouts import VirtualLayout
from ezdxf import colors
from ezdxf.lldxf import const
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.math import Matrix44, Z_AXIS, Vec3, WCSTransform, X_AXIS

from ezdxf.entities.mleader import (
    LeaderLine,
    LeaderData,
    compile_context_tags,
    MLeaderContext,
    MultiLeader,
    BlockData,
    MTextData,
)
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text


@pytest.fixture
def msp():
    return VirtualLayout()


def test_new_multileader(msp):
    """Tests for building a new MULTILEADER in test suite 713."""
    mleader = msp.new_entity("MULTILEADER", {})
    assert mleader.dxftype() == "MULTILEADER"
    assert mleader.dxf.style_handle is None


def test_synonym_mleader(msp):
    mleader = msp.new_entity("MLEADER", {})
    assert mleader.dxftype() == "MLEADER"
    assert mleader.dxf.style_handle is None


class TestMLeaderStyle:
    @pytest.fixture(scope="class")
    def doc(self):
        return ezdxf.new("R2007")

    @pytest.fixture(scope="class")
    def new_style(self, doc):
        return doc.mleader_styles.new("TEST1")

    def test_standard_mleader_style(self, doc):
        """The MLEADERSTYLE entity is more a template, most attributes are stored
        in the MULTILEADER entity itself.

        """
        mleader_style = doc.mleader_styles.get("Standard")
        assert mleader_style.dxftype() == "MLEADERSTYLE"
        assert mleader_style.dxf.content_type == 2
        textstyle_handle = mleader_style.dxf.text_style_handle
        standard = doc.entitydb.get(textstyle_handle)
        assert standard.dxf.name == "Standard"
        assert (
            mleader_style.dxf.leader_linetype_handle is None
        ), "default linetype handle is not set (BYLAYER)"
        assert (
            mleader_style.dxf.arrow_head_handle is None
        ), "default arrow handle is not set (closed filled)"
        assert (
            mleader_style.dxf.block_record_handle is None
        ), "BLOCK_RECORD handle is not set"

    def test_audit_fixes_invalid_text_style_handle(self, doc, new_style):
        new_style.dxf.text_style_handle = "ABBA"
        auditor = Auditor(doc)
        new_style.audit(auditor)
        assert len(auditor.fixes) == 1
        text_style = doc.entitydb.get(new_style.dxf.text_style_handle)
        assert text_style.dxf.name == "Standard"

    def test_audit_fixes_invalid_arrow_head_handle(self, doc, new_style):
        new_style.dxf.arrow_head_handle = "ABBA"
        auditor = Auditor(doc)
        new_style.audit(auditor)
        assert len(auditor.fixes) == 1
        assert new_style.dxf.arrow_head_handle is None, "reset to None"

    def test_audit_fixes_invalid_block_record_handle(self, doc, new_style):
        new_style.dxf.block_record_handle = "ABBA"
        auditor = Auditor(doc)
        new_style.audit(auditor)
        assert len(auditor.fixes) == 1
        assert new_style.dxf.block_record_handle is None, "reset to None"


def matrix(scale=1.0, rotate=0, tx=0, ty=0, tz=0) -> Matrix44:
    return (
        Matrix44.scale(scale)
        @ Matrix44.z_rotate(radians(rotate))
        @ Matrix44.translate(tx, ty, tz)
    )


class TestLeaderLine:
    @pytest.fixture(scope="class")
    def tags(self):
        return Tags.from_text(LEADER_LINE_1)

    def test_parse(self, tags):
        line = LeaderLine.load(tags)
        assert len(line.vertices) == 1
        assert len(line.breaks) == 3
        assert line.index == 0
        assert line.color == colors.BY_BLOCK_RAW_VALUE

    def test_export_dxf(self, tags):
        expected = basic_tags_from_text(LEADER_LINE_1)
        line = LeaderLine.load(tags)
        collector = TagCollector()
        line.export_dxf(collector)
        assert collector.tags == expected

    def test_transform(self):
        point = Vec3(2, 3, 0)
        marker_value = 777
        m = matrix(rotate=30, tx=7, ty=9)
        point_transformed = m.transform(point)

        line = LeaderLine()
        line.vertices.append(point)
        line.breaks.append(point)
        line.index = marker_value
        line.color = marker_value
        line.transform(WCSTransform(m))

        assert line.vertices[0].isclose(point_transformed)
        assert line.breaks[0].isclose(point_transformed)
        assert line.index == marker_value
        assert line.color == marker_value


LEADER_LINE_1 = """304
LEADER_LINE{
 10
181.0
 20
176.0
 30
0.0
 90
0
 11
204.0
 21
159.0
 31
0.0
 12
206.0
 22
158.0
 32
0.0
 91
0
 92
-1056964608
305
}
"""


class TestLeader:
    @pytest.fixture(scope="class")
    def tags(self):
        return Tags.from_text(LEADER_1)

    def test_parse(self, tags):
        ctx = compile_context_tags(tags, 303)
        leader = LeaderData.load(ctx)
        assert len(leader.lines) == 1
        assert leader.has_last_leader_line == 1
        assert leader.has_dogleg_vector == 1
        assert leader.last_leader_point == (213.9, 199.1, 0)
        assert leader.dogleg_vector == (1, 0, 0)
        assert len(leader.breaks) == 2
        assert leader.dogleg_length == 8.0
        assert leader.index == 0

    def test_export_dxf(self, tags):
        expected = basic_tags_from_text(LEADER_1)
        ctx = compile_context_tags(tags, 303)
        leader = LeaderData.load(ctx)
        collector = TagCollector()
        leader.export_dxf(collector)
        assert collector.tags == expected

    def test_transform(self):
        point = Vec3(2, 3, 0)
        marker_value = 777
        dogleg_length = 1.0
        m = matrix(rotate=30, tx=7, ty=9)
        point_transformed = m.transform(point)
        direction_transformed = m.transform_direction(point)

        leader = LeaderData()
        leader.last_leader_point = point
        leader.dogleg_vector = point
        leader.dogleg_length = dogleg_length
        leader.breaks.append(point)
        leader.has_last_leader_line = marker_value
        leader.has_dogleg_vector = marker_value

        leader.transform(WCSTransform(m))
        assert leader.last_leader_point.isclose(point_transformed)
        assert leader.dogleg_vector.isclose(direction_transformed.normalize())
        assert leader.dogleg_length == pytest.approx(dogleg_length)
        assert leader.breaks[0].isclose(point_transformed)
        assert leader.has_last_leader_line == marker_value
        assert leader.has_dogleg_vector == marker_value

    def test_scaling(self):
        dogleg_length = 2.0
        scale = 3.0
        m = matrix(scale=scale, rotate=30, tx=7, ty=9)
        leader = LeaderData()
        leader.dogleg_length = dogleg_length
        leader.transform(WCSTransform(m))
        assert leader.dogleg_length == pytest.approx(dogleg_length * scale)


LEADER_1 = """302
LEADER{
290
1
291
1
10
213.9
20
199.1
30
0.0
11
1.0
21
0.0
31
0.0
12
215.2
22
199.1
32
0.0
13
219.0
23
199.1
33
0.0
90
0
40
8.0
304
LEADER_LINE{
10
195.8
20
176.1
30
0.0
91
0
92
-1056964608
305
}
271
0
303
}
"""


class MLeaderTesting:
    @pytest.fixture(scope="class")
    def tags(self, text):
        tags = Tags.from_text(text)
        return MultiLeader.extract_context_data(tags)

    @pytest.fixture(scope="class")
    def ctx(self, tags):
        return MLeaderContext.load(compile_context_tags(tags, 301))

    @pytest.fixture(scope="class")
    def mleader(self, text):
        return MultiLeader.load(ExtendedTags.from_text(text))

    def test_context_attribs_definition(self, ctx):
        for name in ctx.ATTRIBS.values():
            assert hasattr(ctx, name) is True

    def test_mleader_export_dxf(self, text, mleader):
        expected = basic_tags_from_text(text)
        collector = TagCollector(dxfversion=const.DXF2010)
        mleader.export_dxf(collector)
        assert collector.tags == expected


class TestMTextContext(MLeaderTesting):
    @pytest.fixture(scope="class")
    def text(self):
        return MTEXT_MLEADER_R2010

    def test_mtext_data_attribs_definition(self, ctx):
        mtext = ctx.mtext
        for name in mtext.ATTRIBS.values():
            assert hasattr(mtext, name) is True

    def test_load_mtext_context(self, ctx):
        # Leader() class is tested in TestLeader():
        assert len(ctx.leaders) == 2
        assert ctx.scale == 1
        assert ctx.base_point == (187.4, 185, 0)
        assert ctx.char_height == 5
        assert ctx.arrow_head_size == 3
        assert ctx.landing_gap_size == 2.5
        assert ctx.left_attachment == 1
        assert ctx.right_attachment == 1
        assert ctx.attachment_type == 0
        assert ctx.mtext is not None  # see test_mtext_data()
        assert ctx.block is None
        assert ctx.plane_origin == (1, 2, 3)
        assert ctx.plane_x_axis == (0, 1, 0)
        assert ctx.plane_y_axis == (1, 0, 0)
        assert ctx.plane_normal_reversed == 1
        assert ctx.top_attachment == 8
        assert ctx.bottom_attachment == 8
        assert ctx.plane_z_axis.isclose(Z_AXIS)

    def test_mtext_data(self, ctx):
        mtext = ctx.mtext
        assert mtext.default_content == "MTEXT-DATA-CONTENT"
        assert mtext.extrusion == (1, 0, 0)
        assert mtext.style_handle == "FEFE"  # handle of TextStyle() table entry
        assert mtext.insert == (236.6, 187.0, 0)
        assert mtext.text_direction == (0, 1, 0)
        assert mtext.rotation == 0.2  # in radians!
        assert mtext.width == 104.6
        assert mtext.line_spacing_factor == 1.5
        assert mtext.line_spacing_style == 1
        assert mtext.color == colors.BY_BLOCK_RAW_VALUE
        assert mtext.alignment == 3
        assert mtext.flow_direction == 1
        assert mtext.bg_color == -939524096  # use window background color?
        assert mtext.bg_scale_factor == 2
        assert mtext.bg_transparency == 0
        assert mtext.use_window_bg_color == 0
        assert mtext.has_bg_fill == 0
        assert mtext.column_type == 0
        assert mtext.use_auto_height == 0
        assert mtext.column_width == 0.0
        assert mtext.column_gutter_width == 0.0
        assert mtext.column_flow_reversed == 0
        assert len(mtext.column_sizes) == 0
        assert mtext.use_word_break == 0

    def test_transform_context(self):
        point = Vec3(2, 3, 0)
        marker_value = 777
        length = 1.0
        scale = 3.0
        m = matrix(scale=scale, rotate=30, tx=7, ty=9)
        point_transformed = m.transform(point)

        ctx = MLeaderContext()
        ctx.base_point = point
        ctx.char_height = length
        ctx.arrow_head_size = length
        ctx.landing_gap_size = length
        ctx.left_attachment = marker_value
        ctx.right_attachment = marker_value
        ctx.top_attachment = marker_value
        ctx.bottom_attachment = marker_value
        ctx.text_align_type = marker_value
        ctx.attachment_type = marker_value

        ctx.transform(WCSTransform(m))
        assert ctx.scale == pytest.approx(scale)
        assert ctx.base_point.isclose(point_transformed)
        assert ctx.char_height == pytest.approx(scale * length)
        assert ctx.arrow_head_size == pytest.approx(scale * length)
        assert ctx.landing_gap_size == pytest.approx(scale * length)
        assert ctx.left_attachment == marker_value
        assert ctx.right_attachment == marker_value
        assert ctx.top_attachment == marker_value
        assert ctx.bottom_attachment == marker_value
        assert ctx.text_align_type == marker_value
        assert ctx.attachment_type == marker_value
        assert ctx.plane_origin.isclose(m.transform((0, 0, 0)))
        assert ctx.plane_x_axis.isclose(
            m.transform_direction((1, 0, 0)).normalize()
        )
        assert ctx.plane_y_axis.isclose(
            m.transform_direction((0, 1, 0)).normalize()
        )
        assert ctx.plane_normal_reversed == 0

    def test_transform_context_reversed_extrusion(self):
        ctx = MLeaderContext()
        ctx.transform(WCSTransform(Matrix44.scale(-1, 1, 1)))
        assert ctx.plane_normal_reversed == 1

    def test_transform_mtext_data(self):
        point = Vec3(2, 3, 0)
        length = 1.5
        scale = 3.0
        m = matrix(scale=scale, rotate=30, tx=7, ty=9)
        point_transformed = m.transform(point)

        mtext = MTextData()
        mtext.insert = point
        mtext.width = length
        mtext.defined_height = length
        mtext.column_width = length
        mtext.column_gutter_width = length
        mtext.column_sizes = [length]
        mtext.transform(WCSTransform(m))

        assert mtext.insert.isclose(point_transformed)
        assert mtext.text_direction.isclose(
            m.transform_direction(X_AXIS, normalize=True)
        )
        # rotation in radians!
        assert mtext.rotation == pytest.approx(radians(30))
        assert mtext.width == pytest.approx(scale * length)
        assert mtext.defined_height == pytest.approx(scale * length)
        assert mtext.column_width == pytest.approx(scale * length)
        assert mtext.column_gutter_width == pytest.approx(scale * length)
        assert mtext.column_sizes[0] == pytest.approx(scale * length)

    def test_transform_mtext_extrusion(self):
        """The extrusion vector is always created by the right-hand-rule from
        the transformed x- and y-axis: Z = X "cross" Y.
        """
        mtext = MTextData()
        m = Matrix44.scale(-1, 1, 1)
        mtext.transform(WCSTransform(m))
        assert mtext.text_direction.isclose(m.transform(X_AXIS))
        assert mtext.extrusion.isclose(
            -m.transform(Z_AXIS)
        ), "expected reversed z-axis"


MTEXT_MLEADER_R2010 = """0
MULTILEADER
5
98
330
1F
100
AcDbEntity
8
0
100
AcDbMLeader
270
2
300
CONTEXT_DATA{
40
1.0
10
187.4
20
185.0
30
0.0
41
5.0
140
3.0
145
2.5
174
1
175
1
176
2
177
0
290
1
304
MTEXT-DATA-CONTENT
11
1.0
21
0.0
31
0.0
340
FEFE
12
236.6
22
187.0
32
0.0
13
0.0
23
1.0
33
0.0
42
0.2
43
104.6
44
0.0
45
1.5
170
1
90
-1056964608
171
3
172
1
91
-939524096
141
2.0
92
0
291
0
292
0
173
0
293
0
142
0.0
143
0.0
294
0
295
0
296
0
110
1.0
120
2.0
130
3.0
111
0.0
121
1.0
131
0.0
112
1.0
122
0.0
132
0.0
297
1
302
LEADER{
290
1
291
1
10
246.6
20
185.0
30
0.0
11
-1.0
21
0.0
31
0.0
90
0
40
8.0
304
LEADER_LINE{
10
287.3
20
220.5
30
0.0
91
0
92
-1056964608
305
}
271
0
303
}
302
LEADER{
290
1
291
1
10
179.4
20
185.0
30
0.0
11
1.0
21
0.0
31
0.0
90
1
40
8.0
304
LEADER_LINE{
10
146.5
20
149.0
30
0.0
91
1
92
-1056964608
305
}
271
0
303
}
272
8
273
8
301
}
340
6D
90
330752
170
1
91
-1056964608
341
14
171
-2
290
1
291
1
41
8.0
42
4.0
172
2
343
11
173
1
95
1
174
1
175
0
92
-1056964608
292
0
93
-1056964608
10
1.0
20
1.0
30
1.0
43
0.0
176
0
293
0
294
0
178
0
179
1
45
1.0
271
0
272
9
273
9
"""


class TestBlockContext(MLeaderTesting):
    @pytest.fixture(scope="class")
    def text(self):
        return BLOCK_MLEADER_R2010

    def test_block_data_attribs_definition(self, ctx):
        block = ctx.block
        for name in block.ATTRIBS.values():
            assert hasattr(block, name) is True

    def test_load_block_context(self, ctx):
        # Leader() class is tested in TestLeader():
        assert len(ctx.leaders) == 1
        assert ctx.scale == 1
        assert ctx.base_point == (8.42, 0.70, 0)
        assert ctx.char_height == 5
        assert ctx.arrow_head_size == 3
        assert ctx.landing_gap_size == 2.5
        assert ctx.left_attachment == 1
        assert ctx.right_attachment == 1
        assert ctx.attachment_type == 0
        assert ctx.mtext is None
        assert ctx.block is not None  # see test_block_data()
        assert ctx.plane_origin == (1, 2, 3)
        assert ctx.plane_x_axis == (0, 1, 0)
        assert ctx.plane_y_axis == (1, 0, 0)
        assert ctx.plane_normal_reversed == 1
        assert ctx.top_attachment == 8
        assert ctx.bottom_attachment == 8

    def test_block_data(self, ctx):
        block = ctx.block
        assert block.block_record_handle == "FEFE"
        assert block.extrusion == (0, 0, 1)
        assert block.insert == (18.42, 0.70, 0)
        assert block.scale == (1.0, 2.0, 3.0)
        assert block.rotation == 0.2
        assert block.color == colors.BY_BLOCK_RAW_VALUE

    def test_get_transformation_matrix(self, ctx):
        # The transformation matrix is stored in transposed order
        # of ezdxf.math.Matrix44()!
        # fmt: off
        assert ctx.block._matrix == [
            1, 0, 0, 18.42,
            0, 1, 0, 0.70,
            0, 0, 1, 0,
            0, 0, 0, 1,
        ]
        assert ctx.block.matrix44.get_row(3) == (18.42, 0.70, 0, 1)
        # fmt: on

    def test_set_transformation_matrix(self):
        m = Matrix44()
        m.set_row(3, (4, 3, 2, 1))
        block = BlockData()
        block.matrix44 = m
        # The transformation matrix is stored in transposed order
        # of ezdxf.math.Matrix44()!
        # fmt: off
        assert block._matrix == [
            1, 0, 0, 4,
            0, 1, 0, 3,
            0, 0, 1, 2,
            0, 0, 0, 1,
        ]
        # fmt: on

    def test_transform_block_data(self):
        point = Vec3(2, 3, 0)
        scale = 3.0
        m = matrix(scale=scale, rotate=30, tx=7, ty=9)

        block = BlockData()
        block.insert = point
        block.transform(WCSTransform(m))
        assert block.insert.isclose(m.transform(point))
        assert block.scale.isclose(Vec3(scale, scale, scale))
        assert block.rotation == pytest.approx(radians(30))
        assert list(block.matrix44) == pytest.approx(list(m))

    def test_transform_block_data_x_reflection(self):
        block = BlockData()
        block.transform(WCSTransform(Matrix44.scale(-1, 1, 1)))
        assert block.extrusion.isclose(-Z_AXIS)
        assert block.scale.isclose((1, -1, 1))  # ???

    def test_transform_block_data_y_reflection(self):
        block = BlockData()
        block.transform(WCSTransform(Matrix44.scale(1, -1, 1)))
        assert block.extrusion.isclose(-Z_AXIS)
        assert block.scale.isclose((1, -1, 1))  # ???

    def test_transform_block_data_xy_reflection(self):
        block = BlockData()
        block.transform(WCSTransform(Matrix44.scale(-1, -1, 1)))
        assert block.extrusion.isclose(Z_AXIS)
        assert block.scale.isclose((1, 1, 1))  # ???


BLOCK_MLEADER_R2010 = """  0
MULTILEADER
5
B5
330
1F
100
AcDbEntity
8
0
100
AcDbMLeader
270
2
300
CONTEXT_DATA{
40
1.0
10
8.42
20
0.70
30
0.0
41
5.0
140
3.0
145
2.5
174
1
175
1
176
0
177
0
290
0
296
1
341
FEFE
14
0.0
24
0.0
34
1.0
15
18.42
25
0.70
35
0.0
16
1.0
26
2.0
36
3.0
46
0.2
93
-1056964608
47
1.0
47
0.0
47
0.0
47
18.42
47
0.0
47
1.0
47
0.0
47
0.70
47
0.0
47
0.0
47
1.0
47
0.0
47
0.0
47
0.0
47
0.0
47
1.0
110
1.0
120
2.0
130
3.0
111
0.0
121
1.0
131
0.0
112
1.0
122
0.0
132
0.0
297
1
302
LEADER{
290
1
291
1
10
9.42
20
0.70
30
0.0
11
1.0
21
0.0
31
0.0
90
0
40
8.0
304
LEADER_LINE{
10
1.15
20
-10.40
30
0.0
91
0
92
-1056964608
305
}
271
0
303
}
272
8
273
8
301
}
340
6D
90
6816768
170
1
91
-1056964608
341
14
171
-2
290
1
291
1
41
8.0
42
4.0
172
1
343
11
173
1
95
1
174
1
175
0
92
-1056964608
292
0
344
94
93
-1056964608
10
1.0
20
1.0
30
1.0
43
0.0
176
0
293
0
330
A3
177
1
44
0.0
302
B
294
0
178
0
179
1
45
1.0
271
0
272
9
273
9
"""



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest

from ezdxf.entities import OLE2Frame
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from binascii import unhexlify


@pytest.fixture
def ole2frame():
    return OLE2Frame.from_text(BITMAP)


def test_sublass_is_loaded(ole2frame):
    assert ole2frame.acdb_ole2frame is not None


def test_dxf_type(ole2frame):
    assert ole2frame.dxftype() == "OLE2FRAME"


def test_bounding_box(ole2frame):
    bbox = ole2frame.bbox()
    assert bbox.extmin.isclose((58, 296))
    assert bbox.extmax.isclose((104, 304))


def test_export_exact_original_data(ole2frame):
    expected = basic_tags_from_text(BITMAP)
    collector = TagCollector(dxfversion="AC1015", optional=True)
    ole2frame.export_dxf(collector)
    assert collector.tags == expected, "expected exact same DXF tags"


def test_binary_data(ole2frame):
    expected = unhexlify(
        "8055A6788D2B542F4D401E9880AF920E73400000000000000000078BDDB41A3E"
        "5A401E9880AF920E73400000000000000000078BDDB41A3E5A40938EBE02EC84"
        "72400000000000000000A6788D2B542F4D40938EBE02EC847240000000000000"
    )
    assert ole2frame.binary_data() == expected


BITMAP = """  0
OLE2FRAME
  5
9A
330
1F
100
AcDbEntity
  8
0
100
AcDbOle2Frame
 70
2
  3
Picture (Device Independent Bitmap)
 10
58
 20
304
 30
0.0
 11
104
 21
296
 31
0.0
 71
3
 72
0
 73
0
 90
0
310
8055A6788D2B542F4D401E9880AF920E73400000000000000000078BDDB41A3E
310
5A401E9880AF920E73400000000000000000078BDDB41A3E5A40938EBE02EC84
310
72400000000000000000A6788D2B542F4D40938EBE02EC847240000000000000
  1
OLE
1001
ACAD
1000
OLEBEGIN
1070
70
1070
1
1070
71
1070
1
1070
40
1040
0.0
1070
41
1040
46.600623867295
1070
42
1040
8.603192098578
1070
72
1070
0
1070
3
1000

1070
90
1071
12
1070
43
1040
4.23333
1070
4
1000

1070
91
1071
12
1070
44
1040
4.23333
1000
OLEEND
"""
if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021-2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterable, Iterator
from ezdxf.entities import DXFGraphic, DXFEntity
from ezdxf.lldxf import const
from ezdxf.lldxf.tagwriter import AbstractTagWriter
from ezdxf.protocols import SupportsVirtualEntities
from ezdxf.entities.copy import default_copy, CopyNotSupported


class DXFGraphicProxy(DXFGraphic):
    """DO NOT USE THIS WRAPPER AS REAL DXF ENTITY OUTSIDE THE DRAWING ADD-ON!"""

    def __init__(self, entity: DXFEntity):
        super().__init__()
        self.entity = entity
        self.dxf = self._setup_dxf_namespace(entity)

    def _setup_dxf_namespace(self, entity):
        # copy DXF namespace - modifications do not effect the wrapped entity
        dxf = entity.dxf.copy(self)
        # setup mandatory DXF attributes without default values like layer:
        for k, v in self.DEFAULT_ATTRIBS.items():
            if not dxf.hasattr(k):
                dxf.set(k, v)
        return dxf

    def dxftype(self) -> str:
        return self.entity.dxftype()

    def __virtual_entities__(self) -> Iterator[DXFGraphic]:
        """Implements the SupportsVirtualEntities protocol."""
        if isinstance(self.entity, SupportsVirtualEntities):
            return self.entity.__virtual_entities__()
        if hasattr(self.entity, "virtual_entities"):
            return self.entity.virtual_entities()
        return iter([])

    def virtual_entities(self) -> Iterable[DXFGraphic]:
        return self.__virtual_entities__()

    def copy(self, copy_strategy=default_copy) -> DXFGraphicProxy:
        raise CopyNotSupported(f"Copying of DXFGraphicProxy() not supported.")

    def preprocess_export(self, tagwriter: AbstractTagWriter) -> bool:
        # prevent dxf export
        return False



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
from typing import cast
import pytest
import ezdxf
from ezdxf.entities.idbuffer import IDBuffer
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

IDBUFFER = """0
IDBUFFER
5
0
102
{ACAD_REACTORS
330
0
102
}
330
0
100
AcDbIdBuffer
"""


@pytest.fixture
def entity():
    return IDBuffer.from_text(IDBUFFER)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "IDBUFFER" in ENTITY_CLASSES


def test_default_init():
    entity = IDBuffer()
    assert entity.dxftype() == "IDBUFFER"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = IDBuffer.new(handle="ABBA", owner="0", dxfattribs={})
    assert len(entity.handles) == 0


def test_load_from_text(entity):
    assert len(entity.handles) == 0


def test_write_dxf():
    entity = IDBuffer.from_text(IDBUFFER)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(IDBUFFER)
    assert result == expected


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2007")


def test_generic_id_buffer(doc):
    id_buffer = doc.objects.new_entity("IDBUFFER", {})
    assert id_buffer.dxftype() == "IDBUFFER"
    assert len(id_buffer.handles) == 0


def test_set_get_id_buffer(doc):
    id_buffer = doc.objects.new_entity("IDBUFFER", {})
    assert id_buffer.dxftype() == "IDBUFFER"
    id_buffer.handles = ["FF", "EE", "DD"]
    handles = id_buffer.handles
    assert len(handles) == 3
    assert handles == ["FF", "EE", "DD"]

    handles.append("FFFF")
    assert handles[-1] == "FFFF"

    handles.clear()
    assert len(handles) == 0


def test_dxf_tags(doc):
    id_buffer = cast(IDBuffer, doc.objects.new_entity("IDBUFFER", {}))
    id_buffer.handles = ["FF", "EE", "DD", "CC"]
    tags = TagCollector.dxftags(id_buffer)[-4:]

    assert len(tags) == 4
    assert tags[0] == (330, "FF")
    assert tags[-1] == (330, "CC")


def test_clone(doc):
    id_buffer = cast(IDBuffer, doc.objects.new_entity("IDBUFFER", {}))
    id_buffer.handles = ["FF", "EE", "DD", "CC"]
    buffer2 = cast(IDBuffer, id_buffer.copy())
    buffer2.handles[-1] = "ABCD"
    assert id_buffer.handles[:-1] == buffer2.handles[:-1]
    assert id_buffer.handles[-1] != buffer2.handles[-1]



# Copyright (c) 2020-2023, Matthew Broadway
# License: MIT License
# mypy: ignore_errors=True
from __future__ import annotations
from typing import Optional, Iterable
import abc
import math

import numpy as np

from ezdxf.addons.xqt import QtCore as qc, QtGui as qg, QtWidgets as qw
from ezdxf.addons.drawing.backend import Backend, BkPath2d, BkPoints2d, ImageData
from ezdxf.addons.drawing.config import Configuration
from ezdxf.addons.drawing.type_hints import Color
from ezdxf.addons.drawing.properties import BackendProperties
from ezdxf.math import Vec2, Matrix44
from ezdxf.npshapes import to_qpainter_path


class _Point(qw.QAbstractGraphicsShapeItem):
    """A dimensionless point which is drawn 'cosmetically' (scale depends on
    view)
    """

    def __init__(self, x: float, y: float, brush: qg.QBrush):
        super().__init__()
        self.location = qc.QPointF(x, y)
        self.radius = 1.0
        self.setPen(qg.QPen(qc.Qt.NoPen))
        self.setBrush(brush)

    def paint(
        self,
        painter: qg.QPainter,
        option: qw.QStyleOptionGraphicsItem,
        widget: Optional[qw.QWidget] = None,
    ) -> None:
        view_scale = _get_x_scale(painter.transform())
        radius = self.radius / view_scale
        painter.setBrush(self.brush())
        painter.setPen(qc.Qt.NoPen)
        painter.drawEllipse(self.location, radius, radius)

    def boundingRect(self) -> qc.QRectF:
        return qc.QRectF(self.location, qc.QSizeF(1, 1))


# The key used to store the dxf entity corresponding to each graphics element
CorrespondingDXFEntity = qc.Qt.UserRole + 0  # type: ignore
CorrespondingDXFParentStack = qc.Qt.UserRole + 1  # type: ignore


class _PyQtBackend(Backend):
    """
    Abstract PyQt backend which uses the :mod:`PySide6` package to implement an
    interactive viewer. The :mod:`PyQt5` package can be used as fallback if the
    :mod:`PySide6` package is not available.
    """

    def __init__(self, scene: qw.QGraphicsScene):
        super().__init__()
        self._scene = scene
        self._color_cache: dict[Color, qg.QColor] = {}
        self._no_line = qg.QPen(qc.Qt.NoPen)
        self._no_fill = qg.QBrush(qc.Qt.NoBrush)

    def configure(self, config: Configuration) -> None:
        if config.min_lineweight is None:
            config = config.with_changes(min_lineweight=0.24)
        super().configure(config)

    def set_scene(self, scene: qw.QGraphicsScene) -> None:
        self._scene = scene

    def _add_item(self, item: qw.QGraphicsItem, entity_handle: str) -> None:
        self.set_item_data(item, entity_handle)
        self._scene.addItem(item)

    @abc.abstractmethod
    def set_item_data(self, item: qw.QGraphicsItem, entity_handle: str) -> None:
        ...

    def _get_color(self, color: Color) -> qg.QColor:
        try:
            return self._color_cache[color]
        except KeyError:
            pass
        if len(color) == 7:
            qt_color = qg.QColor(color)  # '#RRGGBB'
        elif len(color) == 9:
            rgb = color[1:7]
            alpha = color[7:9]
            qt_color = qg.QColor(f"#{alpha}{rgb}")  # '#AARRGGBB'
        else:
            raise TypeError(color)

        self._color_cache[color] = qt_color
        return qt_color

    def _get_pen(self, properties: BackendProperties) -> qg.QPen:
        """Returns a cosmetic pen with applied lineweight but without line type
        support.
        """
        px = properties.lineweight / 0.3527 * self.config.lineweight_scaling
        pen = qg.QPen(self._get_color(properties.color), px)
        # Use constant width in pixel:
        pen.setCosmetic(True)
        pen.setJoinStyle(qc.Qt.RoundJoin)
        return pen

    def _get_fill_brush(self, color: Color) -> qg.QBrush:
        return qg.QBrush(self._get_color(color), qc.Qt.SolidPattern)  # type: ignore

    def set_background(self, color: Color):
        self._scene.setBackgroundBrush(qg.QBrush(self._get_color(color)))

    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        """Draw a real dimensionless point."""
        brush = self._get_fill_brush(properties.color)
        item = _Point(pos.x, pos.y, brush)
        self._add_item(item, properties.handle)

    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        # PyQt draws a long line for a zero-length line:
        if start.isclose(end):
            self.draw_point(start, properties)
        else:
            item = qw.QGraphicsLineItem(start.x, start.y, end.x, end.y)
            item.setPen(self._get_pen(properties))
            self._add_item(item, properties.handle)

    def draw_solid_lines(
        self,
        lines: Iterable[tuple[Vec2, Vec2]],
        properties: BackendProperties,
    ):
        """Fast method to draw a bunch of solid lines with the same properties."""
        pen = self._get_pen(properties)
        add_line = self._add_item
        for s, e in lines:
            if s.isclose(e):
                self.draw_point(s, properties)
            else:
                item = qw.QGraphicsLineItem(s.x, s.y, e.x, e.y)
                item.setPen(pen)
                add_line(item, properties.handle)

    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        if len(path) == 0:
            return
        item = qw.QGraphicsPathItem(to_qpainter_path([path]))
        item.setPen(self._get_pen(properties))
        item.setBrush(self._no_fill)
        self._add_item(item, properties.handle)

    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        # Default fill rule is OddEvenFill! Detecting the path orientation is not
        # necessary!
        _paths = list(paths)
        if len(_paths) == 0:
            return
        item = _CosmeticPath(to_qpainter_path(_paths))
        item.setPen(self._get_pen(properties))
        item.setBrush(self._get_fill_brush(properties.color))
        self._add_item(item, properties.handle)

    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        brush = self._get_fill_brush(properties.color)
        polygon = qg.QPolygonF()
        for p in points.vertices():
            polygon.append(qc.QPointF(p.x, p.y))
        item = _CosmeticPolygon(polygon)
        item.setPen(self._no_line)
        item.setBrush(brush)
        self._add_item(item, properties.handle)

    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        image = image_data.image
        transform = image_data.transform
        height, width, depth = image.shape
        assert depth == 4
        bytes_per_row = width * depth
        image = np.ascontiguousarray(np.flip(image, axis=0))
        pixmap = qg.QPixmap(
            qg.QImage(
                image.data,
                width,
                height,
                bytes_per_row,
                qg.QImage.Format.Format_RGBA8888,
            )
        )
        item = qw.QGraphicsPixmapItem()
        item.setPixmap(pixmap)
        item.setTransformationMode(qc.Qt.TransformationMode.SmoothTransformation)
        item.setTransform(_matrix_to_qtransform(transform))
        self._add_item(item, properties.handle)

    def clear(self) -> None:
        self._scene.clear()

    def finalize(self) -> None:
        super().finalize()
        self._scene.setSceneRect(self._scene.itemsBoundingRect())


class PyQtBackend(_PyQtBackend):
    """
    Backend which uses the :mod:`PySide6` package to implement an interactive
    viewer. The :mod:`PyQt5` package can be used as fallback if the :mod:`PySide6`
    package is not available.

    Args:
        scene: drawing canvas of type :class:`QtWidgets.QGraphicsScene`,
            if ``None`` a new canvas will be created
    """

    def __init__(
        self,
        scene: Optional[qw.QGraphicsScene] = None,
    ):
        super().__init__(scene or qw.QGraphicsScene())

    # This implementation keeps all virtual entities alive by attaching references
    # to entities to the graphic scene items.

    def set_item_data(self, item: qw.QGraphicsItem, entity_handle: str) -> None:
        parent_stack = tuple(e for e, props in self.entity_stack[:-1])
        current_entity = self.current_entity
        item.setData(CorrespondingDXFEntity, current_entity)
        item.setData(CorrespondingDXFParentStack, parent_stack)


class PyQtPlaybackBackend(_PyQtBackend):
    """
    Backend which uses the :mod:`PySide6` package to implement an interactive
    viewer. The :mod:`PyQt5` package can be used as fallback if the :mod:`PySide6`
    package is not available.

    This backend can be used a playback backend for the :meth:`replay` method of the
    :class:`Player` class

    Args:
        scene: drawing canvas of type :class:`QtWidgets.QGraphicsScene`,
            if ``None`` a new canvas will be created
    """

    def __init__(
        self,
        scene: Optional[qw.QGraphicsScene] = None,
    ):
        super().__init__(scene or qw.QGraphicsScene())

    # The backend recorder does not record enter_entity() and exit_entity() events.
    # This implementation attaches only entity handles (str) to the graphic scene items.
    # Each item references the top level entity e.g. all items of a block reference
    # references the handle of the INSERT entity.

    def set_item_data(self, item: qw.QGraphicsItem, entity_handle: str) -> None:
        item.setData(CorrespondingDXFEntity, entity_handle)


class _CosmeticPath(qw.QGraphicsPathItem):
    def paint(
        self,
        painter: qg.QPainter,
        option: qw.QStyleOptionGraphicsItem,
        widget: Optional[qw.QWidget] = None,
    ) -> None:
        _set_cosmetic_brush(self, painter)
        super().paint(painter, option, widget)


class _CosmeticPolygon(qw.QGraphicsPolygonItem):
    def paint(
        self,
        painter: qg.QPainter,
        option: qw.QStyleOptionGraphicsItem,
        widget: Optional[qw.QWidget] = None,
    ) -> None:
        _set_cosmetic_brush(self, painter)
        super().paint(painter, option, widget)


def _set_cosmetic_brush(
    item: qw.QAbstractGraphicsShapeItem, painter: qg.QPainter
) -> None:
    """like a cosmetic pen, this sets the brush pattern to appear the same independent of the view"""
    brush = item.brush()
    # scale by -1 in y because the view is always mirrored in y and undoing the view transformation entirely would make
    # the hatch mirrored w.r.t the view
    brush.setTransform(painter.transform().inverted()[0].scale(1, -1))  # type: ignore
    item.setBrush(brush)


def _get_x_scale(t: qg.QTransform) -> float:
    return math.sqrt(t.m11() * t.m11() + t.m21() * t.m21())


def _matrix_to_qtransform(matrix: Matrix44) -> qg.QTransform:
    """Qt also uses row-vectors so the translation elements are placed in the
    bottom row.

    This is only a simple conversion which assumes that although the
    transformation is 4x4,it does not involve the z axis.

    A more correct transformation could be implemented like so:
    https://stackoverflow.com/questions/10629737/convert-3d-4x4-rotation-matrix-into-2d
    """
    return qg.QTransform(*matrix.get_2d_transformation())



#  Copyright (c) 2022-2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import TypeVar, TYPE_CHECKING
import abc
from ezdxf.fonts import fonts

if TYPE_CHECKING:
    from ezdxf.npshapes import NumpyPath2d

T = TypeVar("T")


class TextRenderer(abc.ABC):
    """Minimal requirement to be usable as a universal text renderer"""

    @abc.abstractmethod
    def get_font_measurements(
        self, font_face: fonts.FontFace, cap_height: float = 1.0
    ) -> fonts.FontMeasurements:
        ...

    @abc.abstractmethod
    def get_text_line_width(
        self,
        text: str,
        font_face: fonts.FontFace,
        cap_height: float = 1.0,
    ) -> float:
        ...

    @abc.abstractmethod
    def get_text_path(
        self, text: str, font_face: fonts.FontFace, cap_height: float = 1.0
    ) -> NumpyPath2d:
        ...

    @abc.abstractmethod
    def get_text_glyph_paths(
        self, text: str, font_face: fonts.FontFace, cap_height: float = 1.0
    ) -> list[NumpyPath2d]:
        ...

    @abc.abstractmethod
    def is_stroke_font(self, font_face: fonts.FontFace) -> bool:
        ...



# Copyright (c) 2019-2023 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Sequence, Optional, Iterator
from ezdxf.lldxf.types import dxftag, uniform_appid
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.const import DXFKeyError, DXFStructureError
from ezdxf.lldxf.const import (
    ACAD_REACTORS,
    REACTOR_HANDLE_CODE,
    APP_DATA_MARKER,
)

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

__all__ = ["AppData", "Reactors"]

ERR_INVALID_DXF_ATTRIB = "Invalid DXF attribute for entity {}"
ERR_DXF_ATTRIB_NOT_EXITS = "DXF attribute {} does not exist"


class AppData:
    def __init__(self) -> None:
        self.data: dict[str, Tags] = dict()

    def __contains__(self, appid: str) -> bool:
        """Returns ``True`` if application-defined data exist for `appid`."""
        return uniform_appid(appid) in self.data

    def __len__(self) -> int:
        """Returns the count of AppData."""
        return len(self.data)

    def tags(self) -> Iterable[Tags]:
        return self.data.values()

    def get(self, appid: str) -> Tags:
        """Get application-defined data for `appid` as
        :class:`~ezdxf.lldxf.tags.Tags` container.
        The first tag is always (102, "{APPID").
        The last tag is always (102, "}").
        """
        try:
            return self.data[uniform_appid(appid)]
        except KeyError:
            raise DXFKeyError(appid)

    def set(self, tags: Tags) -> None:
        """Store raw application-defined data tags.
        The first tag has to be (102, "{APPID").
        The last tag has to be (102, "}").
        """
        if len(tags):
            appid = tags[0].value
            self.data[appid] = tags

    def add(self, appid: str, data: Iterable[Sequence]) -> None:
        """Add application-defined tags for `appid`.
        Adds first tag (102, "{APPID") if not exist.
        Adds last tag (102, "}" if not exist.
        """
        data = Tags(dxftag(code, value) for code, value in data)
        appid = uniform_appid(appid)
        if data[0] != (APP_DATA_MARKER, appid):
            data.insert(0, dxftag(APP_DATA_MARKER, appid))
        if data[-1] != (APP_DATA_MARKER, "}"):
            data.append(dxftag(APP_DATA_MARKER, "}"))
        self.set(data)

    def discard(self, appid: str):
        """Delete application-defined data for `appid` without raising and error
        if `appid` doesn't exist.
        """
        _appid = uniform_appid(appid)
        if _appid in self.data:
            del self.data[_appid]

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        for data in self.data.values():
            tagwriter.write_tags(data)


class Reactors:
    """Handle storage for related reactors.

    Reactors are other objects related to the object that contains this
    Reactor() instance.

    """

    def __init__(self, handles: Optional[Iterable[str]] = None):
        self.reactors: set[str] = set(handles or [])

    def __len__(self) -> int:
        """Returns count of registered handles."""
        return len(self.reactors)

    def __contains__(self, handle: str) -> bool:
        """Returns ``True`` if `handle` is registered."""
        return handle in self.reactors

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator for all registered handles."""
        return iter(self.get())

    def copy(self) -> Reactors:
        """Returns a copy."""
        return Reactors(self.reactors)

    @classmethod
    def from_tags(cls, tags: Optional[Tags] = None) -> Reactors:
        """Create Reactors() instance from tags.

        Expected DXF structure:
        [(102, '{ACAD_REACTORS'), (330, handle), ..., (102, '}')]

        Args:
            tags: list of DXFTags()

        """
        if tags is None:
            return cls(None)

        if len(tags) < 2:  # no reactors are valid
            raise DXFStructureError("ACAD_REACTORS error")
        return cls((handle.value for handle in tags[1:-1]))

    def get(self) -> list[str]:
        """Returns all registered handles as sorted list."""
        return sorted(self.reactors, key=lambda x: int(x, base=16))

    def set(self, handles: Optional[Iterable[str]]) -> None:
        """Reset all handles."""
        self.reactors = set(handles or [])

    def add(self, handle: str) -> None:
        """Add a single `handle`."""
        self.reactors.add(handle)

    def discard(self, handle: str):
        """Discard a single `handle`."""
        self.reactors.discard(handle)

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        tagwriter.write_tag2(APP_DATA_MARKER, ACAD_REACTORS)
        for handle in self.get():
            tagwriter.write_tag2(REACTOR_HANDLE_CODE, handle)
        tagwriter.write_tag2(APP_DATA_MARKER, "}")



# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Iterator
from ezdxf.math import Vec2, Shape2d, NULLVEC, UVec
from .forms import open_arrow, arrow2

if TYPE_CHECKING:
    from ezdxf.entities import DXFGraphic
    from ezdxf.sections.blocks import BlocksSection
    from ezdxf.eztypes import GenericLayoutType

DEFAULT_ARROW_ANGLE = 18.924644
DEFAULT_BETA = 45.0


# The base arrow is oriented for the right hand side ->| of the dimension line,
# reverse is the left hand side |<-.
class BaseArrow:
    def __init__(self, vertices: Iterable[UVec]):
        self.shape = Shape2d(vertices)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        pass

    def place(self, insert: UVec, angle: float):
        self.shape.rotate(angle)
        self.shape.translate(insert)


class NoneStroke(BaseArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        super().__init__([Vec2(insert)])


class ObliqueStroke(BaseArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        self.size = size
        s2 = size / 2
        # shape = [center, lower left, upper right]
        super().__init__([Vec2((-s2, -s2)), Vec2((s2, s2))])
        self.place(insert, angle)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        layout.add_line(
            start=self.shape[0], end=self.shape[1], dxfattribs=dxfattribs
        )


class ArchTick(ObliqueStroke):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        width = self.size * 0.15
        dxfattribs = dxfattribs or {}
        if layout.dxfversion > "AC1009":
            dxfattribs["const_width"] = width
            layout.add_lwpolyline(
                self.shape, format="xy", dxfattribs=dxfattribs  # type: ignore
            )
        else:
            dxfattribs["default_start_width"] = width
            dxfattribs["default_end_width"] = width
            layout.add_polyline2d(self.shape, dxfattribs=dxfattribs)  # type: ignore


class ClosedArrowBlank(BaseArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        super().__init__(open_arrow(size, angle=DEFAULT_ARROW_ANGLE))
        self.place(insert, angle)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        if layout.dxfversion > "AC1009":
            polyline = layout.add_lwpolyline(
                points=self.shape, dxfattribs=dxfattribs  # type: ignore
            )
        else:
            polyline = layout.add_polyline2d(  # type: ignore
                points=self.shape, dxfattribs=dxfattribs  # type: ignore
            )
        polyline.close(True)


class ClosedArrow(ClosedArrowBlank):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        super().render(layout, dxfattribs)
        end_point = self.shape[0].lerp(self.shape[2])

        layout.add_line(
            start=self.shape[1], end=end_point, dxfattribs=dxfattribs
        )


class ClosedArrowFilled(ClosedArrow):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        layout.add_solid(
            points=self.shape,  # type: ignore
            dxfattribs=dxfattribs,
        )


class _OpenArrow(BaseArrow):
    def __init__(
        self,
        arrow_angle: float,
        insert: UVec,
        size: float = 1.0,
        angle: float = 0,
    ):
        points = list(open_arrow(size, angle=arrow_angle))
        points.append((-1, 0))
        super().__init__(points)
        self.place(insert, angle)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        if layout.dxfversion > "AC1009":
            layout.add_lwpolyline(points=self.shape[:-1], dxfattribs=dxfattribs)
        else:
            layout.add_polyline2d(points=self.shape[:-1], dxfattribs=dxfattribs)
        layout.add_line(
            start=self.shape[1], end=self.shape[-1], dxfattribs=dxfattribs
        )


class OpenArrow(_OpenArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        super().__init__(DEFAULT_ARROW_ANGLE, insert, size, angle)


class OpenArrow30(_OpenArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        super().__init__(30, insert, size, angle)


class OpenArrow90(_OpenArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        super().__init__(90, insert, size, angle)


class Circle(BaseArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        self.radius = size / 2
        # shape = [center point, connection point]
        super().__init__(
            [
                Vec2((0, 0)),
                Vec2((-self.radius, 0)),
                Vec2((-size, 0)),
            ]
        )
        self.place(insert, angle)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        layout.add_circle(
            center=self.shape[0], radius=self.radius, dxfattribs=dxfattribs
        )


class Origin(Circle):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        super().render(layout, dxfattribs)
        layout.add_line(
            start=self.shape[0], end=self.shape[2], dxfattribs=dxfattribs
        )


class CircleBlank(Circle):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        super().render(layout, dxfattribs)
        layout.add_line(
            start=self.shape[1], end=self.shape[2], dxfattribs=dxfattribs
        )


class Origin2(Circle):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        layout.add_circle(
            center=self.shape[0], radius=self.radius, dxfattribs=dxfattribs
        )
        layout.add_circle(
            center=self.shape[0], radius=self.radius / 2, dxfattribs=dxfattribs
        )
        layout.add_line(
            start=self.shape[1], end=self.shape[2], dxfattribs=dxfattribs
        )


class DotSmall(Circle):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        center = self.shape[0]
        d = Vec2((self.radius / 2, 0))
        p1 = center - d
        p2 = center + d
        dxfattribs = dxfattribs or {}
        if layout.dxfversion > "AC1009":
            dxfattribs["const_width"] = self.radius
            layout.add_lwpolyline(
                [(p1, 1), (p2, 1)],
                format="vb",
                close=True,
                dxfattribs=dxfattribs,
            )
        else:
            dxfattribs["default_start_width"] = self.radius
            dxfattribs["default_end_width"] = self.radius
            polyline = layout.add_polyline2d(
                points=[p1, p2], close=True, dxfattribs=dxfattribs
            )
            polyline[0].dxf.bulge = 1
            polyline[1].dxf.bulge = 1


class Dot(DotSmall):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        layout.add_line(
            start=self.shape[1], end=self.shape[2], dxfattribs=dxfattribs
        )
        super().render(layout, dxfattribs)


class Box(BaseArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        # shape = [lower_left, lower_right, upper_right, upper_left, connection point]
        s2 = size / 2
        super().__init__(
            [
                Vec2((-s2, -s2)),
                Vec2((+s2, -s2)),
                Vec2((+s2, +s2)),
                Vec2((-s2, +s2)),
                Vec2((-s2, 0)),
                Vec2((-size, 0)),
            ]
        )
        self.place(insert, angle)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        if layout.dxfversion > "AC1009":
            polyline = layout.add_lwpolyline(
                points=self.shape[0:4], dxfattribs=dxfattribs
            )
        else:
            polyline = layout.add_polyline2d(  # type: ignore
                points=self.shape[0:4], dxfattribs=dxfattribs
            )
        polyline.close(True)
        layout.add_line(
            start=self.shape[4], end=self.shape[5], dxfattribs=dxfattribs
        )


class BoxFilled(Box):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        def solid_order():
            v = self.shape.vertices
            return [v[0], v[1], v[3], v[2]]

        layout.add_solid(points=solid_order(), dxfattribs=dxfattribs)
        layout.add_line(
            start=self.shape[4], end=self.shape[5], dxfattribs=dxfattribs
        )


class Integral(BaseArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        self.radius = size * 0.3535534
        self.angle = angle
        # shape = [center, left_center, right_center]
        super().__init__(
            [
                Vec2((0, 0)),
                Vec2((-self.radius, 0)),
                Vec2((self.radius, 0)),
            ]
        )
        self.place(insert, angle)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        angle = self.angle
        layout.add_arc(
            center=self.shape[1],
            radius=self.radius,
            start_angle=-90 + angle,
            end_angle=angle,
            dxfattribs=dxfattribs,
        )
        layout.add_arc(
            center=self.shape[2],
            radius=self.radius,
            start_angle=90 + angle,
            end_angle=180 + angle,
            dxfattribs=dxfattribs,
        )


class DatumTriangle(BaseArrow):
    REVERSE_ANGLE = 180

    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        d = 0.577350269 * size  # tan(30)
        # shape = [upper_corner, lower_corner, connection_point]
        super().__init__(
            [
                Vec2((0, d)),
                Vec2((0, -d)),
                Vec2((-size, 0)),
            ]
        )
        self.place(insert, angle)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        if layout.dxfversion > "AC1009":
            polyline = layout.add_lwpolyline(
                points=self.shape, dxfattribs=dxfattribs  # type: ignore
            )
        else:
            polyline = layout.add_polyline2d(  # type: ignore
                points=self.shape, dxfattribs=dxfattribs  # type: ignore
            )
        polyline.close(True)


class DatumTriangleFilled(DatumTriangle):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        layout.add_solid(points=self.shape, dxfattribs=dxfattribs)  # type: ignore


class _EzArrow(BaseArrow):
    def __init__(self, insert: UVec, size: float = 1.0, angle: float = 0):
        points = list(arrow2(size, angle=DEFAULT_ARROW_ANGLE))
        points.append((-1, 0))
        super().__init__(points)
        self.place(insert, angle)

    def render(self, layout: GenericLayoutType, dxfattribs=None):
        if layout.dxfversion > "AC1009":
            polyline = layout.add_lwpolyline(
                self.shape[:-1], dxfattribs=dxfattribs
            )
        else:
            polyline = layout.add_polyline2d(  # type: ignore
                self.shape[:-1], dxfattribs=dxfattribs
            )
        polyline.close(True)


class EzArrowBlank(_EzArrow):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        super().render(layout, dxfattribs)
        layout.add_line(
            start=self.shape[-2], end=self.shape[-1], dxfattribs=dxfattribs
        )


class EzArrow(_EzArrow):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        super().render(layout, dxfattribs)
        layout.add_line(
            start=self.shape[1], end=self.shape[-1], dxfattribs=dxfattribs
        )


class EzArrowFilled(_EzArrow):
    def render(self, layout: GenericLayoutType, dxfattribs=None):
        points = self.shape.vertices
        layout.add_solid(
            [points[0], points[1], points[3], points[2]], dxfattribs=dxfattribs
        )
        layout.add_line(
            start=self.shape[-2], end=self.shape[-1], dxfattribs=dxfattribs
        )


class _Arrows:
    closed_filled = ""
    dot = "DOT"
    dot_small = "DOTSMALL"
    dot_blank = "DOTBLANK"
    origin_indicator = "ORIGIN"
    origin_indicator_2 = "ORIGIN2"
    open = "OPEN"
    right_angle = "OPEN90"
    open_30 = "OPEN30"
    closed = "CLOSED"
    dot_smallblank = "SMALL"
    none = "NONE"
    oblique = "OBLIQUE"
    box_filled = "BOXFILLED"
    box = "BOXBLANK"
    closed_blank = "CLOSEDBLANK"
    datum_triangle_filled = "DATUMFILLED"
    datum_triangle = "DATUMBLANK"
    integral = "INTEGRAL"
    architectural_tick = "ARCHTICK"
    # ezdxf special arrows
    ez_arrow = "EZ_ARROW"
    ez_arrow_blank = "EZ_ARROW_BLANK"
    ez_arrow_filled = "EZ_ARROW_FILLED"

    CLASSES = {
        closed_filled: ClosedArrowFilled,
        dot: Dot,
        dot_small: DotSmall,
        dot_blank: CircleBlank,
        origin_indicator: Origin,
        origin_indicator_2: Origin2,
        open: OpenArrow,
        right_angle: OpenArrow90,
        open_30: OpenArrow30,
        closed: ClosedArrow,
        dot_smallblank: Circle,
        none: NoneStroke,
        oblique: ObliqueStroke,
        box_filled: BoxFilled,
        box: Box,
        closed_blank: ClosedArrowBlank,
        datum_triangle: DatumTriangle,
        datum_triangle_filled: DatumTriangleFilled,
        integral: Integral,
        architectural_tick: ArchTick,
        ez_arrow: EzArrow,
        ez_arrow_blank: EzArrowBlank,
        ez_arrow_filled: EzArrowFilled,
    }
    # arrows with origin at dimension line start/end
    ORIGIN_ZERO = {
        architectural_tick,
        oblique,
        dot_small,
        dot_smallblank,
        integral,
        none,
    }

    __acad__ = {
        closed_filled,
        dot,
        dot_small,
        dot_blank,
        origin_indicator,
        origin_indicator_2,
        open,
        right_angle,
        open_30,
        closed,
        dot_smallblank,
        none,
        oblique,
        box_filled,
        box,
        closed_blank,
        datum_triangle,
        datum_triangle_filled,
        integral,
        architectural_tick,
    }
    __ezdxf__ = {
        ez_arrow,
        ez_arrow_blank,
        ez_arrow_filled,
    }
    __all_arrows__ = __acad__ | __ezdxf__

    EXTENSIONS_ALLOWED = {
        architectural_tick,
        oblique,
        none,
        dot_smallblank,
        integral,
        dot_small,
    }

    def is_acad_arrow(self, item: str) -> bool:
        """Returns ``True`` if `item` is a standard AutoCAD arrow."""
        return item.upper() in self.__acad__

    def is_ezdxf_arrow(self, item: str) -> bool:
        """Returns ``True`` if `item` is a special `ezdxf` arrow."""
        return item.upper() in self.__ezdxf__

    def has_extension_line(self, name):
        """Returns ``True`` if the arrow `name` supports extension lines."""
        return name in self.EXTENSIONS_ALLOWED

    def __contains__(self, item: str) -> bool:
        """Returns `True` if `item` is an arrow managed by this class."""
        if item is None:
            return False
        return item.upper() in self.__all_arrows__

    def create_block(self, blocks: BlocksSection, name: str) -> str:
        """Creates the BLOCK definition for arrow `name`."""
        block_name = self.block_name(name)
        if block_name not in blocks:
            block = blocks.new(block_name)
            arrow = self.arrow_shape(name, insert=(0, 0), size=1, rotation=0)
            arrow.render(block, dxfattribs={"color": 0, "linetype": "BYBLOCK"})
        return block_name

    def arrow_handle(self, blocks: BlocksSection, name: str) -> str:
        """Returns the BLOCK_RECORD handle for arrow `name`."""
        arrow_name = self.arrow_name(name)
        block_name = self.create_block(blocks, arrow_name)
        block = blocks.get(block_name)
        return block.block_record_handle

    def block_name(self, name: str) -> str:
        """Returns the block name."""
        if not self.is_acad_arrow(name):  # common BLOCK definition
            # e.g. Dimension.dxf.bkl = 'EZ_ARROW' == Insert.dxf.name
            return name.upper()
        elif name == "":
            # special AutoCAD arrow symbol 'CLOSED_FILLED' has no name
            # ezdxf uses blocks for ALL arrows, but '_' (closed filled) as block name?
            return "_CLOSEDFILLED"  # Dimension.dxf.bkl = '' != Insert.dxf.name = '_CLOSED_FILLED'
        else:
            # add preceding '_' to AutoCAD arrow symbol names
            # Dimension.dxf.bkl = 'DOT' != Insert.dxf.name = '_DOT'
            return "_" + name.upper()

    def arrow_name(self, block_name: str) -> str:
        """Returns the arrow name."""
        if block_name.startswith("_"):
            name = block_name[1:].upper()
            if name == "CLOSEDFILLED":
                return ""
            elif self.is_acad_arrow(name):
                return name
        return block_name

    def insert_arrow(
        self,
        layout: GenericLayoutType,
        name: str,
        insert: UVec = NULLVEC,
        size: float = 1.0,
        rotation: float = 0,
        *,
        dxfattribs=None,
    ) -> Vec2:
        """Insert arrow as block reference into `layout`."""
        block_name = self.create_block(layout.doc.blocks, name)

        dxfattribs = dict(dxfattribs or {})
        dxfattribs["rotation"] = rotation
        dxfattribs["xscale"] = size
        dxfattribs["yscale"] = size
        layout.add_blockref(block_name, insert=insert, dxfattribs=dxfattribs)
        return connection_point(
            name, insert=insert, scale=size, rotation=rotation
        )

    def render_arrow(
        self,
        layout: GenericLayoutType,
        name: str,
        insert: UVec = NULLVEC,
        size: float = 1.0,
        rotation: float = 0,
        *,
        dxfattribs=None,
    ) -> Vec2:
        """Render arrow as basic DXF entities into `layout`."""
        dxfattribs = dict(dxfattribs or {})
        arrow = self.arrow_shape(name, insert, size, rotation)
        arrow.render(layout, dxfattribs)
        return connection_point(
            name, insert=insert, scale=size, rotation=rotation
        )

    def virtual_entities(
        self,
        name: str,
        insert: UVec = NULLVEC,
        size: float = 0.625,
        rotation: float = 0,
        *,
        dxfattribs=None,
    ) -> Iterator[DXFGraphic]:
        """Returns all arrow components as virtual DXF entities."""
        from ezdxf.layouts import VirtualLayout

        if name in self:
            layout = VirtualLayout()
            ARROWS.render_arrow(
                layout,
                name,
                insert=insert,
                size=size,
                rotation=rotation,
                dxfattribs=dxfattribs,
            )
            yield from iter(layout)

    def arrow_shape(
        self, name: str, insert: UVec, size: float, rotation: float
    ) -> BaseArrow:
        """Returns an instance of the shape management class for arrow `name`."""
        # size depending shapes
        name = name.upper()
        if name == self.dot_small:
            size *= 0.25
        elif name == self.dot_smallblank:
            size *= 0.5
        cls = self.CLASSES[name]
        return cls(insert, size, rotation)


def connection_point(
    arrow_name: str, insert: UVec, scale: float = 1.0, rotation: float = 0.0
) -> Vec2:
    """Returns the connection point for `arrow_name`."""
    insert = Vec2(insert)
    if ARROWS.arrow_name(arrow_name) in _Arrows.ORIGIN_ZERO:
        return insert
    else:
        return insert - Vec2.from_deg_angle(rotation, scale)


def arrow_length(arrow_name: str, scale: float = 1.0) -> float:
    """Returns the scaled arrow length of `arrow_name`."""
    if ARROWS.arrow_name(arrow_name) in _Arrows.ORIGIN_ZERO:
        return 0.0
    else:
        return scale


ARROWS: _Arrows = _Arrows()



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License

import ezdxf
import pytest

from ezdxf.render.dimension import LinearDimension, DimStyleOverride
from ezdxf.render.dim_base import compile_mtext
from ezdxf.math import Vec3


@pytest.fixture(scope="module")
def dwg():
    return ezdxf.new("R2007", setup=True)


def test_linear_dimension_with_one_tolerance(dwg):
    msp = dwg.modelspace()
    dimline = msp.add_linear_dim(base=(0, 10), p1=(0, 0), p2=(100, 0))
    override = {
        "dimlfac": 1,
        "dimtol": 1,
        "dimtfac": 0.5,
        "dimtolj": 0,
        "dimtp": 0.01,
        "dimtm": 0.01,
    }
    style = DimStyleOverride(dimline.dimension, override)
    renderer = LinearDimension(dimline.dimension, override=style)
    assert renderer.measurement.text == "100"
    assert renderer.measurement.decimal_separator == "."
    assert renderer.tol.decimal_places == 2  # default value
    assert renderer.tol.text == "±0.01"
    assert renderer.tol.valign == 0
    assert (
        compile_mtext(renderer.measurement, renderer.tol) == r"\A0;100{\H0.50x;±0.01}"
    )


def test_linear_dimension_with_two_tolerances(dwg):
    msp = dwg.modelspace()
    dimline = msp.add_linear_dim(base=(0, 10), p1=(0, 0), p2=(101, 0))
    override = {
        "dimlfac": 1,
        "dimtol": 1,
        "dimtfac": 0.5,
        "dimtolj": 1,
        "dimtp": 0.02,
        "dimtm": 0.03,
    }
    style = DimStyleOverride(dimline.dimension, override)
    renderer = LinearDimension(dimline.dimension, override=style)
    assert renderer.measurement.text == "101"
    assert renderer.measurement.decimal_separator == "."
    assert renderer.tol.decimal_places == 2  # default value
    assert renderer.tol.text_upper == "+0.02"
    assert renderer.tol.text_lower == "-0.03"
    assert renderer.tol.valign == 1
    assert (
        compile_mtext(renderer.measurement, renderer.tol)
        == r"\A1;101{\H0.50x;\S+0.02^ -0.03;}"
    )


def test_linear_dimension_with_limits(dwg):
    msp = dwg.modelspace()
    dimline = msp.add_linear_dim(base=(0, 10), p1=(0, 0), p2=(101, 0))
    override = {
        "dimlfac": 1,
        "dimlim": 1,
        "dimtfac": 0.5,
        "dimtp": 0.02,
        "dimtm": 0.03,
    }
    style = DimStyleOverride(dimline.dimension, override)
    renderer = LinearDimension(dimline.dimension, override=style)
    assert renderer.measurement.text == "101"
    assert renderer.measurement.decimal_separator == "."
    assert renderer.tol.decimal_places == 2  # default value
    assert renderer.tol.text_upper == "101.02"
    assert renderer.tol.text_lower == "100.97"
    assert (
        compile_mtext(renderer.measurement, renderer.tol)
        == r"{\H0.50x;\S101.02^ 100.97;}"
    )


def test_dimension_insert_attribute_translates_the_block_content():
    doc = ezdxf.new()
    msp = doc.modelspace()
    dimline = msp.add_linear_dim(base=(0, 10), p1=(0, 0), p2=(100, 0))
    dimline.render()
    dim = dimline.dimension
    blk = dim.get_geometry_block()

    blk_points = blk.query("POINT")
    virtual_points = [e for e in dim.virtual_entities() if e.dxftype() == "POINT"]

    # without the insert attribute, the virtual points should be located at the
    # original location:
    for vpoint, blk_point in zip(virtual_points, blk_points):
        assert vpoint.dxf.location.isclose(blk_point.dxf.location)

    # set an insertion point for this dimension ...
    INSERT = Vec3(10, 10, 0)
    dim.dxf.insert = Vec3(INSERT)
    virtual_points = [e for e in dim.virtual_entities() if e.dxftype() == "POINT"]

    # ... and the virtual points should be translated by the insert vector
    for vpoint, blk_point in zip(virtual_points, blk_points):
        assert (vpoint.dxf.location - blk_point.dxf.location).isclose(INSERT)


@pytest.mark.parametrize("color", [1, 7])
def test_override_all_colors(color):
    new_doc = ezdxf.new()
    new_msp = new_doc.modelspace()
    style_override = {"dimclrt": color, "dimclrd": color, "dimclre": color}
    dim_renderer = new_msp.add_linear_dim(
        base=(0, 10), p1=(0, 0), p2=(100, 0), override=style_override
    ).render()
    dimension = dim_renderer.dimension
    blk = dimension.get_geometry_block()
    for entity in blk:
        if entity.dxftype() == "POINT":
            continue
        assert entity.dxf.color == color



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
import pytest

import ezdxf
from ezdxf.layouts import Modelspace
from ezdxf import select


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new()
    msp_ = doc.modelspace()

    msp_.add_point((0, 1))
    msp_.add_circle((0, 0), radius=5)
    msp_.add_line((-1, -1), (1, 1))
    msp_.add_lwpolyline([(-2, -2), (2, -2), (2, 2), (-2, 2)], close=True)
    return msp_


class TestBoundingBoxVsWindow:
    """The selection functions are testing only the bounding boxes of entities.

    This is a design choice: performance and simplicity over accuracy
    """

    def test_all_inside(self, msp: Modelspace):
        window = select.Window((-5, -5), (5, 5))
        selection = select.bbox_inside(window, msp)
        assert len(selection) == 4

    def test_inside_select_point(self, msp: Modelspace):
        window = select.Window((-1, 0), (1, 2))
        selection = select.bbox_inside(window, msp)
        assert len(selection) == 1
        assert selection[0].dxftype() == "POINT"

    def test_inside_select_point_and_line(self, msp: Modelspace):
        window = select.Window((-1, -1), (1, 1))
        selection = select.bbox_inside(window, msp)
        assert len(selection) == 2
        assert selection[0].dxftype() == "POINT"
        assert selection[1].dxftype() == "LINE"

    def test_outside_none(self, msp: Modelspace):
        window = select.Window((-5, -5), (5, 5))
        selection = select.bbox_outside(window, msp)
        assert len(selection) == 0

    def test_outside_all(self, msp: Modelspace):
        window = select.Window((6, 6), (7, 7))
        selection = select.bbox_outside(window, msp)
        assert len(selection) == 4

    def test_overlaps_all(self, msp: Modelspace):
        """The window just overlaps the bounding boxes of the CIRCLE and the
        LWPOLYLINE not the geometry itself.
        """
        window = select.Window((0, 0), (2, 2))
        selection = select.bbox_overlap(window, msp)
        assert len(selection) == 4

    def test_overlap_selects_touching_entities(self, msp: Modelspace):
        """The window touches just the bounding box of the CIRCLE, not the curve itself."""
        window = select.Window((5, 5), (6, 6))
        selection = select.bbox_overlap(window, msp)
        assert len(selection) == 1
        assert selection[0].dxftype() == "CIRCLE"


class TestBoundingBoxVsCircle:
    """The selection functions are testing only the bounding boxes of selection.

    This is a design choice: performance and simplicity over accuracy
    """

    def test_all_inside(self, msp: Modelspace):
        circle = select.Circle((0, 0), 10)
        selection = select.bbox_inside(circle, msp)
        assert len(selection) == 4

    def test_all_outside(self, msp: Modelspace):
        circle = select.Circle((7, 7), 1)
        selection = select.bbox_outside(circle, msp)
        assert len(selection) == 4

    def test_none_outside(self, msp: Modelspace):
        circle = select.Circle((0, 0), 1)
        selection = select.bbox_outside(circle, msp)
        assert len(selection) == 0

    def test_outside_corner_case_point(self, msp: Modelspace):
        """Bounding boxes do overlap but point is outside circle.

        point = (0, 1)
        """
        circle = select.Circle((1, 0), 1)
        selection = select.bbox_outside(circle, msp.query("POINT"))
        assert len(selection) == 1

    def test_outside_corner_case_polyline(self, msp: Modelspace):
        """All vertices are outside circle but polyline is not outside.

        xCCx
        CCCC
        CCCC
        xCCx

        polyline corners = (-2, -2) ... (2, 2)
        """
        circle = select.Circle((0, 0), 2)
        selection = select.bbox_outside(circle, msp.query("LWPOLYLINE"))
        assert len(selection) == 0

    def test_overlap_all(self, msp: Modelspace):
        circle = select.Circle((0, 0), 1)
        selection = select.bbox_overlap(circle, msp)
        assert len(selection) == 4

    def test_overlap_none(self, msp: Modelspace):
        circle = select.Circle((7, 0), 1)
        selection = select.bbox_overlap(circle, msp)
        assert len(selection) == 0

    def test_is_overlapping_point(self, msp: Modelspace):
        circle = select.Circle((0, 0), 1)
        selection = select.bbox_overlap(circle, msp.query("POINT"))
        assert len(selection) == 1

    def test_is_overlapping_line(self, msp: Modelspace):
        """Bounding boxes do overlap but circle does not intersect line.
        overlap tests are performed on the bounding box of the line!

        CC.
        C#x
        .xx

        line = (-1, -1) (1, 1)
        """
        circle = select.Circle((-1, 1), 1)
        selection = select.bbox_overlap(circle, msp.query("LINE"))
        assert len(selection) == 1


class TestBoundingBoxVsPolygon:
    """The selection functions are testing only the bounding boxes of selection.

    This is a design choice: performance and simplicity over accuracy
    """

    @pytest.mark.parametrize(
        "vertices",
        [
            [],
            [(0, 0)],
            [(0, 0), (0, 0)],
            [(0, 0), (0, 0), (0, 0)],
        ],
    )
    def test_invalid_vertex_count(self, vertices):
        with pytest.raises(ValueError):
            select.Polygon(vertices)

    def test_all_inside(self, msp: Modelspace):
        polygon = select.Polygon([(-5, -5), (5, -5), (5, 5), (-5, 5)])
        selection = select.bbox_inside(polygon, msp)
        assert len(selection) == 4

    def test_circle_not_inside(self, msp: Modelspace):
        """Bounding boxes do overlap but all corner vertices of the CIRCLE bbox are
        outside.
        """
        polygon = select.Polygon([(-2, -2), (2, -2), (2, 2), (-2, 2)])
        selection = select.bbox_inside(polygon, msp.query("CIRCLE"))
        assert len(selection) == 0

    def test_circle_is_not_inside_concave_polygon(self, msp: Modelspace):
        """Bounding box of CIRCLE is completely inside the polygon, all edge vertices
        are inside the polygon but some edges do intersect with the polygon.
        CIRCLE is not complete inside!
        """
        polygon = select.Polygon([(-10, -10), (10, -10), (10, 10), (0, 0), (-10, 10)])
        selection = select.bbox_inside(polygon, msp.query("CIRCLE"))
        assert len(selection) == 0

        selection = select.bbox_outside(polygon, msp.query("CIRCLE"))
        assert len(selection) == 0

    def test_all_outside(self, msp: Modelspace):
        polygon = select.Polygon([(6, 6), (7, 6), (7, 7), (6, 7)])
        selection = select.bbox_outside(polygon, msp)
        assert len(selection) == 4

    def test_point_is_outside(self, msp: Modelspace):
        """Bounding boxes do overlap, all corner vertices of the POINT bbox are
        outside and POINT is outside the polygon.

        point: (0, 1)
        """
        polygon = select.Polygon([(0, 0), (1, 0), (1, 1)])
        selection = select.bbox_outside(polygon, msp.query("POINT"))
        assert len(selection) == 1

    def test_line_is_not_outside(self, msp: Modelspace):
        """Bounding boxes do overlap all corner vertices of the LINE bbox are
        outside but the LINE is overlap the polygon.

        line: (-2, -2) (2, 2)
        """
        polygon = select.Polygon([(-1, -3), (1, -3), (0, 3)])
        selection = select.bbox_outside(polygon, msp.query("LINE"))
        assert len(selection) == 0

    def test_all_overlap(self, msp: Modelspace):
        """All inside is also all overlap."""
        polygon = select.Polygon([(-5, -5), (5, -5), (5, 5), (-5, 5)])
        selection = select.bbox_overlap(polygon, msp)
        assert len(selection) == 4

    def test_point_not_overlaps(self, msp: Modelspace):
        """Bounding boxes do overlap, all corner vertices of the POINT bbox are
        outside and POINT is outside the polygon.

        point: (0, 1)
        """
        polygon = select.Polygon([(0, 0), (1, 0), (1, 1)])
        selection = select.bbox_overlap(polygon, msp.query("POINT"))
        assert len(selection) == 0

    def test_line_overlaps(self, msp: Modelspace):
        """Bounding boxes do overlap, all corner vertices of the LINE bbox are
        outside and the LINE is overlap the polygon.

        line: (-2, -2) (2, 2)
        """
        polygon = select.Polygon([(-1, -3), (1, -3), (0, 3)])
        selection = select.bbox_overlap(polygon, msp.query("LINE"))
        assert len(selection) == 1

    def test_circle_overlaps(self, msp: Modelspace):
        """All corner vertices of the CIRCLE bbox are outside and CIRCLE does overlap
        with polygon.  This is different to overlap selection in CAD applications!

        circle: (0, 0) radius=5
        """
        polygon = select.Polygon([(-1, -1), (1, -1), (1, 1), (-1, 1)])
        selection = select.bbox_overlap(polygon, msp.query("CIRCLE"))
        assert len(selection) == 1

    def test_concave(self, msp: Modelspace):
        polygon = select.Polygon(
            [(-10, -10), (-10, -20), (20, -20), (20, 20), (10, 20), (10, -10)]
        )
        # ...PP
        # .O.PP
        # ...PP
        # PPPPP
        # PPPPP
        selection = select.bbox_inside(polygon, msp.query("CIRCLE"))
        assert len(selection) == 0

        selection = select.bbox_outside(polygon, msp.query("CIRCLE"))
        assert len(selection) == 1

        selection = select.bbox_overlap(polygon, msp.query("CIRCLE"))
        assert len(selection) == 0


class TestBoundingBoxIntersectsFence:
    @pytest.mark.parametrize("vertices", [[], [(0, 0)]])
    def test_invalid_vertex_count(self, vertices, msp: Modelspace):
        with pytest.raises(ValueError):
            select.bbox_crosses_fence(vertices, msp)

    def test_overlap_all_except_point(self, msp: Modelspace):
        selection = select.bbox_crosses_fence([(-5, 0), (5, 0)], msp)
        assert len(selection) == 3

    def test_can_not_select_point(self, msp: Modelspace):
        """Fence cannot select POINT entities by definition.

        point: (0, 1)
        """
        selection = select.bbox_crosses_fence([(-5, 1), (5, 1)], msp.query("POINT"))
        assert len(selection) == 0

    def test_select_nothing_inside_a_closed_fence(self, msp: Modelspace):
        """A closed fence does NOT work like a polygon selection!"""
        selection = select.bbox_crosses_fence(
            [(-9, -9), (-9, 9), (9, 9), (-9, 9), (-9, -9)], msp
        )
        assert len(selection) == 0

    def test_select_by_touching_bbox_corner(self, msp: Modelspace):
        """CIRCLE is selected by fence touching the CIRCLE bbox at one corner point.
        Does not cross the curve itself!

        circle: (0, 0), radius = 5
        """
        selection = select.bbox_crosses_fence([(0, 10), (10, 0)], msp)
        assert len(selection) == 1


def test_point_selects_all(msp: Modelspace):
    """The point is inside of all entity bounding boxes."""
    selection = select.point_in_bbox((0, 1), msp)
    assert len(selection) == 4


def test_all_entities_chained_by_bounding_boxes():
    doc = ezdxf.new()
    msp = doc.modelspace()
    for x in range(10):
        msp.add_line((x, 0), (x + 1.01, 1))

    start = msp[-1]
    selected = select.bbox_chained(start, msp)
    assert len(selected) == 10


class TestPlanarSearchIndex:
    def test_circle_select_radius_1(self, msp: Modelspace):
        psi = select.PlanarSearchIndex(msp)
        result = psi.detection_point_in_circle((0, 0), 1)
        assert len(result) == 1
        assert result[0].dxftype() == "POINT"
        # CIRCLE is not selected because, none of the bounding box vertices of the
        # CIRCLE is inside the section circle

    def test_circle_select_radius_2(self, msp: Modelspace):
        psi = select.PlanarSearchIndex(msp)
        result = psi.detection_point_in_circle((0, 0), 2)
        assert len(result) == 2
        types = {e.dxftype() for e in result}
        assert "POINT" in types
        assert "LINE" in types

    def test_circle_select_all(self, msp: Modelspace):
        psi = select.PlanarSearchIndex(msp)
        result = psi.detection_point_in_circle((0, 0), 8)
        assert len(result) == 4

    def test_rect_select(self, msp: Modelspace):
        psi = select.PlanarSearchIndex(msp)
        result = psi.detection_point_in_rect((-1, -1), (1, 1))
        assert len(result) == 2
        types = {e.dxftype() for e in result}
        assert "POINT" in types
        assert "LINE" in types

    def test_rect_select_all(self, msp: Modelspace):
        psi = select.PlanarSearchIndex(msp)
        result = psi.detection_point_in_rect((-5, -5), (5, 5))
        assert len(result) == 4


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2011-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Sequence, TYPE_CHECKING, Iterable
from ezdxf.lldxf.tagger import internal_tag_compiler

if TYPE_CHECKING:
    from ezdxf.lldxf.types import DXFTag
    from ezdxf.lldxf.extendedtags import ExtendedTags


def compile_tags_without_handles(text: str) -> Iterable[DXFTag]:
    return (
        tag for tag in internal_tag_compiler(text) if tag.code not in (5, 105)
    )


def normlines(text: str) -> Sequence[str]:
    lines = text.split("\n")
    return [line.strip() for line in lines]


def load_section(text: str, name: str) -> list[ExtendedTags]:
    from ezdxf.lldxf.loader import load_dxf_structure

    dxf = load_dxf_structure(
        internal_tag_compiler(text), ignore_missing_eof=True
    )
    return dxf[name]  # type: ignore


def load_entities(text: str, name: str):
    from ezdxf.lldxf.loader import load_dxf_structure, load_dxf_entities

    dxf = load_dxf_structure(
        internal_tag_compiler(text), ignore_missing_eof=True
    )
    return load_dxf_entities(dxf[name])  # type: ignore


def parse_hex_dump(txt: str) -> bytes:
    b = bytearray()
    lines = txt.split("\n")
    for line in lines:
        if line == "":
            continue
        data = [int(v, 16) for v in line.strip().split(" ")]
        assert data[0] == len(b)
        b.extend(data[1:])
    return b
